#/*
# * Copyright (c) 2019,2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import struct
import sys
import types
import unittest
import os
import getopt
import re
import subprocess
import shutil
from pathlib import Path
from pathlib import PurePath
from io import StringIO
import contextlib
import importlib
from importlib.machinery import SourceFileLoader
import tempfile
# from enum import Enum
import atexit
import textwrap
from collections import UserDict
from collections import OrderedDict

from lopper.fmt import LopperFmt
import lopper.base
import lopper.log
from lopper.base import lopper_base
from lopper.tree import LopperTreePrinter

import lopper.schema

lopper.log._init(__name__)
lopper.log._init("fdt.py")

from string import printable

try:
    import libfdt
    from libfdt import Fdt, FdtException, QUIET_NOTFOUND, QUIET_ALL
except:
    import site
    python_version_dir = f"python{sys.version_info[0]}.{sys.version_info[1]}"
    site.addsitedir((Path(__file__).parent / f'vendor/lib/{python_version_dir}/site-packages').resolve())

    import libfdt
    from libfdt import Fdt, FdtException, QUIET_NOTFOUND, QUIET_ALL

# general retry count
MAX_RETRIES = 10

class LopperFDT(lopper.base.lopper_base):
    """The Lopper Class contains static methods for manipulating device trees

    Use the lopper methods when manipulating device trees (in particular
    libfdt FDT objects) or SystemDeviceTree classes.

    """

    @staticmethod
    def fdt_copy( fdt ):
        """Copy a fdt

        Creats a new FDT that is a copy of the passed one.

        Args:
            fdt (FDT): reference FDT

        Returns:
            fdt: The newly created FDT
        """

        return Fdt( fdt.as_bytearray() )

    @staticmethod
    def fdt( size=None, other_fdt=None ):
        """Create a new FDT

        Creats a new FDT of a passed size. If other_fdt is passed, it
        is used as the start size of the fdt.

        If no size or other fdt is passed, 128 bytes is the default
        size

        Args:
            size (int,optional): size in bytes of the FDT
            other_fdt (FDT,optional): reference FDT for size

        Returns:
            fdt: The newly created FDT
        """

        fdt = None

        if other_fdt:
            size = other_fdt.totalsize()
        else:
            if not size:
                # size is in bytes
                size = 128

            fdt = libfdt.Fdt.create_empty_tree( size )

        return fdt

    @staticmethod
    def node_getname( fdt, node_number_or_path ):
        """Gets the FDT name of a node

        Args:
            fdt (fdt): flattened device tree object
            node_number_or_path: node number or path

        Returns:
            string: name of the node, or "" if node wasn't found
        """
        node_number = -1
        node_path = ""
        try:
            node_number = int(node_number_or_path)
            node_path = LopperFDT.node_abspath( fdt, node_number_or_path )
        except ValueError:
            node_number = LopperFDT.node_find( fdt, node_number_or_path )
            node_path = node_number_or_path
        try:
            name = fdt.get_name( node_number )
        except:
            name = ""

        return name

    @staticmethod
    def node_setname( fdt, node_number_or_path, newname ):
        """Sets the FDT name of a node

        Args:
            fdt (fdt): flattened device tree object
            node_number_or_path: node number or path
            newname (string): name of the node

        Returns:
            boolean: True if the name was set, False otherwise
        """
        node_number = -1
        node_path = ""
        try:
            node_number = int(node_number_or_path)
            node_path = LopperFDT.node_abspath( fdt, node_number_or_path )
        except ValueError:
            node_number = LopperFDT.node_find( fdt, node_number_or_path )
            node_path = node_number_or_path

        retval = False
        if node_number == -1:
            return retval

        for _ in range(MAX_RETRIES):
            try:
                fdt.set_name( node_number, newname )
                retval = True
            except Exception as e:
                fdt.resize( fdt.totalsize() + 1024 )
                continue
            else:
                break

        return retval


    @staticmethod
    def node_find( fdt, node_prefix ):
        """Finds a node by its prefix

        Args:
            fdt (fdt): flattened device tree object
            node_prefix (string): device tree path

        Returns:
            int: node number if successful, otherwise -1
        """
        try:
            node = fdt.path_offset( node_prefix )
            # libfdt returns matches based on partial paths. i.e. if a tree
            # has /timer@2a810000 and /timer, and you search for /timer, you
            # will get the node number for /timer@2a810000 .. this causes problems.
            # so we double check the name, and if it isn't as expected, we do a
            # slower secondary search to find the proper node.
            node_name = fdt.get_name( node )
            if node_name != os.path.basename( node_prefix ):
                node, matching_nodes = LopperFDT.node_find_by_name( fdt, os.path.basename(node_prefix) + "$" )

        except:
            node = -1

        return node

    @staticmethod
    def node_type( fdt, node_offset, verbose=0 ):
        """Utility function to get the "type" of a node

        A small wrapper around the compatible property, but we can use this
        instead of directly getting compatible, since if we switch formats or if
        we want to infer anything based on the name of a node, we can hide it in
        this routine

        Args:
            fdt (fdt): flattened device tree object
            node_offset (int): node number
            verbose (int): verbose output level

        Returns:
            string: compatible string of the node if successful, otherwise ''
        """
        rt = LopperFDT.property_get( fdt, node_offset, "compatible" )

        return rt

    @staticmethod
    def node_by_phandle( fdt, phandle, verbose=0 ):
        """Get a node offset by a phandle

        Thin wrapper around the libfdt routine. The wrapper provides
        consistent exception handling and verbosity level handling.

        Args:
            fdt (fdt): flattened device tree object
            phandle(int): phandle to use as lookup key
            verbose(bool,optional): verbosity level. Deafult is 0.

        Returns:
            int: if > 0, the node that was found. -1 if node was not found.
        """
        anode = -1
        try:
            anode = fdt.node_offset_by_phandle( phandle )
        except:
            pass

        return anode

    @staticmethod
    def node_find_by_name( fdt, node_name, starting_node = 0, multi_match=False ):
        """Finds a node by its name (not path)

        Searches for a node by its name, and returns the offset of that same node
        Note: use this when you don't know the full path of a node

        Args:
            fdt (fdt): flattened device tree object
            node_name (string): name of the node
            starting_node (int): node number to use as the search starting point
            multi_match (bool,optional): flag to indicate if more than one matching
                                         node should be found, default is False

        Returns:
            tuple: first matching node, list of matching nodes. -1 and [] if no match is found
        """

        nn = starting_node
        matching_nodes = []
        matching_node = -1
        # short circuit the search if they are looking for /
        if node_name == "/":
            depth = -1
        else:
            depth = 0

        while depth >= 0:
            nn_name = fdt.get_name(nn)
            if nn_name:
                # we search to support regex matching
                if re.search( node_name, nn_name ):
                    matching_nodes.append(nn)
                    if matching_node == -1:
                        # this is the first match, so we capture the number
                        matching_node = nn

                    if not multi_match:
                        depth = -1
                    else:
                        # match, but mult-match is on .. get the next node
                        nn, depth = fdt.next_node(nn, depth, (libfdt.BADOFFSET,))
                else:
                    # no match, get the next node
                    nn, depth = fdt.next_node(nn, depth, (libfdt.BADOFFSET,))
            else:
                # no name, get the next node
                nn, depth = fdt.next_node(nn, depth, (libfdt.BADOFFSET,))

        return matching_node, matching_nodes

    @staticmethod
    def node_find_by_regex( fdt, node_regex, starting_node = 0, multi_match=False, paths_not_numbers=False ):
        """Finds a node by a regex /path/<regex>/<name>

        Searches for nodes that match a regex (path + name).

        Note: if you pass the name of a node as the regex, you'll get a list of
              that node + children
        Note: if you pass no regex, you'll get all nodes from the starting point
              to the end of the tree.

        Args:
            fdt (fdt): flattened device tree object
            node_regex (string): regex to use for comparision
            starting_node (int): node number to use as the search starting point
            multi_match (bool,optional): flag to indicate if more than one matching
                                         node should be found, default is False
            paths_not_numbers (bool,optional): flag to request paths, not node numbers
                                               be returned

        Returns:
            tuple: first matching node, list of matching nodes. -1 and [] if no match is found
        """

        nn = starting_node
        matching_nodes = []
        matching_node = -1
        depth = 0

        while depth >= 0:
            nn_name = fdt.get_name(nn)
            node_path = ""
            if nn > 0:
                node_path = LopperFDT.node_abspath( fdt, nn )

            if nn_name:
                # we search to support regex matching
                if re.search( node_regex, node_path ):
                    matching_nodes.append(nn)
                    if matching_node == -1:
                        # this is the first match, so we capture the number
                        matching_node = nn

                    if not multi_match:
                        depth = -1
                    else:
                        # match, but mult-match is on .. get the next node
                        nn, depth = fdt.next_node(nn, depth, (libfdt.BADOFFSET,))
                else:
                    # no match, get the next node
                    nn, depth = fdt.next_node(nn, depth, (libfdt.BADOFFSET,))
            else:
                # no name, get the next node
                nn, depth = fdt.next_node(nn, depth, (libfdt.BADOFFSET,))

        # convert everything to paths if requested. This could have been in the
        # loop, but let's keep it simple :D
        if paths_not_numbers:
            matching_node = LopperFDT.node_abspath( fdt, matching_node )
            matching_node_list = matching_nodes
            matching_nodes = []
            for m in matching_node_list:
                matching_nodes.append( LopperFDT.node_abspath( fdt, m ) )

        return matching_node, matching_nodes

    @staticmethod
    def node_prop_check( fdt, node_name, property_name ):
        """Check if a node contains a property

        Boolean check to see if a node contains a property.

        The node name does not need to be a full path or path prefix, since
        the node will be searched starting at the root node, which means that
        a non-unique node name could match multiple nodes.

        Args:
            fdt (fdt): flattened device tree object
            node_name (string): name of the node
            property_name (string): name of the property to check

        Returns:
            bool: True if the node has the property, otherwise False
        """

        node = LopperFDT.node_find( fdt, node_name )
        if node == -1:
            node, nodes = LopperFDT.node_find_by_name( fdt, node_name )

        if node == -1:
            return False

        try:
            fdt.getprop( node, property_name )
        except:
            return False

        return True

    # A thin wrapper + consistent logging and error handling around FDT's
    # node delete
    @staticmethod
    def node_remove( fdt, target_node_offset, verbose = 0 ):
        """remove a node from the device tree

        Thin wrapper and consistent logging around libfdt's node delete.

        Args:
           fdt (fdt): flattended device tree
           target_node_offset (int): offset of the node to be deleted

        Returns:
           Boolean: True if node is removed, false otherwise

        """
        target_node_name = fdt.get_name( target_node_offset )

        lopper.log._debug( f"node_remove: deleting node {target_node_name}" )

        try:
            fdt.del_node( target_node_offset, True )
        except:
            return False

        return True


    @staticmethod
    def node_add( fdt_dest, node_full_path, create_parents = True, verbose = 0 ):
        """Add an empty node to a flattened device tree

        Creates a new node in a flattened devide tree at a given path. If
        desired a node structure (aka parents) will be created as part of
        adding the node at the specified path.

        Args:
            fdt_dest (fdt): flattened device tree object
            node_full_path (string): fully specified path (and name) of the node to create
            create_parents (bool,optional): Should parent nodes be created. Default is True.
                True: create parents as required, False: error if parents are missing
            verbose (int,optional): verbosity level. default is 0.

        Returns:
            int: The node offset of the created node, if successfull, otherwise -1
        """

        lopper.log._debug( f"node_add: creating {node_full_path}" )

        paths_to_check = [ node_full_path ]
        n_path = node_full_path

        # create an ascending list of parent path components to check for existence
        while n_path != "/":
            n_path = os.path.dirname(n_path)
            paths_to_check.insert( 0, n_path )

        # walk that list, create what is missing and use the node numbers as parent offsets
        node_parent = 0
        for p in paths_to_check:
            node_number = LopperFDT.node_find( fdt_dest, p )
            node_name = os.path.basename( p )
            if node_number == -1:
                # were we the last item in the paths to check ? if not, we have
                # to check to see if the parent create flag was set .. if not, return
                # -1 and exit. Otherwise, create all missing components
                if p != paths_to_check[-1]:
                    if not create_parents:
                        lopper.log._debug( f"node_add: parent node {p} missing and parent creation disabled; returning -1" )
                        return -1

                # add it
                for _ in range(MAX_RETRIES):
                    try:
                        lopper.log._debug( f"node_add: adding node '{node_name}' with parent offset {node_parent}" )
                        node_parent = fdt_dest.add_subnode( node_parent, node_name )
                    except FdtException as e:
                        if e.err == -2:
                            lopper.log._debug( "node_add: existing node found, creating temporary placeholder node" )

                            # node exists. This shouldn't happen, unless we've
                            # hit the prefix bug. Let's delete the node that
                            # occupied the slot, add ours, and then re-add the node

                            #node_to_remove_add_readd = fdt.path_offset( node_prefix )

                            # we exploit the fact that the holding node starting
                            # wth loppper<> means that the output export routines
                            # will ignore it later. And as such, we create the
                            # node, and the add properties later create it, so
                            # we get the proper node and don't have to do the
                            # delete -> add -> delete dance. If this breaks, we
                            # may have to revisit and do the extra manipulations.
                            node_name = f"lopper{str(_)}" + node_name
                        else:
                            fdt_dest.resize( fdt_dest.totalsize() + 1024 )

                        continue
                    else:
                        break
            else:
                # it exists
                node_parent = node_number

        return node_parent

    @staticmethod
    def node_properties( fdt, node_number_or_path ):
        """Get the list of properties for a node

        Gather the list of FDT properties for a given node.

        Args:
            fdt (fdt): flattened device tree object
            node_number_or_path: (string or int): node number or full path to
                                 the target node.

        Returns:
            list (FDT prop): The properties of the node [] if no props
        """
        prop_list = []
        node = LopperFDT.node_number( fdt, node_number_or_path )
        if node == -1:
            return prop_list

        poffset = fdt.first_property_offset(node, QUIET_NOTFOUND)
        while poffset > 0:
            prop = fdt.get_property_by_offset(poffset)
            prop_list.append(prop)
            poffset = fdt.next_property_offset(poffset, QUIET_NOTFOUND)

        return prop_list

    @staticmethod
    def nodes( fdt, node_number_or_path, abs_paths = True ):
        """Get the nodes of a tree from a starting point

        Gather the list nodes in the tree from a particular starting point

        Args:
            fdt (fdt): flattened device tree object
            node_number_or_path: (string or int): node number or full path to
                                 the target node.
            abs_paths (boolean, optional): indicate if absolute paths should be returned

        Returns:
            list (strings): The nodes, [] if no nodes
        """
        node_list = []

        node = LopperFDT.node_number( fdt, node_number_or_path )
        if node == -1:
            return node_list

        depth = 0
        while depth >= 0:
            if abs_paths:
                node_list.append( LopperFDT.node_abspath( fdt, node ) )
            else:
                node_list.append( LopperFDT.node_getname( fdt, node ) )

            node, depth = fdt.next_node(node, depth, (libfdt.BADOFFSET,))

        return node_list

    @staticmethod
    def node_subnodes( fdt, node_number_or_path, abs_paths = True ):
        """Get the list of properties for a node

        Gather the list of FDT properties for a given node.

        Args:
            fdt (fdt): flattened device tree object
            node_number_or_path: (string or int): node number or full path to
                                 the target node.
            abs_paths (boolean, optional): indicate if absolute paths should be returned

        Returns:
            list (strings): The subnodes, [] if no subnodes
        """
        node_list = []
        node = LopperFDT.node_number( fdt, node_number_or_path )
        if node == -1:
            return node_list

        offset = fdt.first_subnode(node, QUIET_NOTFOUND)
        while offset > 0:
            if abs_paths:
                node_list.append( LopperFDT.node_abspath( fdt, offset ) )
            else:
                node_list.append( LopperFDT.node_getname( fdt, offset ) )

            offset = fdt.next_subnode( offset, QUIET_NOTFOUND )

        return node_list

    @staticmethod
    def node_parent( fdt, node_number_or_path ):
        """Get the parent offset / number of a node

        Args:
            fdt (fdt): flattened device tree object
            node_number_or_path: (string or int): node number or full path to
                                 the target node.

        Returns:
            int: the node number of the parent
        """
        parent = -1
        node_number = LopperFDT.node_number( fdt, node_number_or_path )
        if node_number > 0:
            parent = fdt.parent_offset( node_number, QUIET_NOTFOUND )

        return parent

    @staticmethod
    def node_sync( fdt, node_in, parent = None, verbose = False ):
        """Write a node description to a FDT

        This routine takes an input dictionary, and writes the details to
        the passed fdt.

        The dictionary contains a set of internal properties, as well as
        a list of standand properties to the node. Internal properties have
        a __ suffix and __ prefix.

        In particular:
            - __path__ : is the absolute path fo the node, and is used to lookup
                         the target node
            - __fdt_name__ : is the name of the node and will be written to the
                             fdt name property
            - __fdt_phandle__ : is the phandle for the node

        All other '/' leading, or '__' leading properties will be written to
        the FDT as node properties.

        If the node doesn't exist, it will be created. If the node exists, then
        the existing properties are read, and any that are no present in the
        passed dictionary are deleted.

        Args:
            fdt (fdt): flattened device tree object
            node_in: (dictionary): Node description dictionary
            parent (string,optional): path to the parent node
            verbose (bool,optional): verbosity level

        Returns:
            Nothing
        """
        lopper.log._debug( f"node_sync: start {node_in['__fdt_name__']} ({node_in['__path__']})")

        nn = LopperFDT.node_find( fdt, node_in['__path__'] )
        if nn == -1:
            # -1 means the node wasn't found
            lopper.log._debug( f"node_sync: creating node {node_in['__path__']}" )

            nn = LopperFDT.node_add( fdt, node_in['__path__'], True )
            if nn == -1:
                lopper.log._error( "node_sync: node could not be added; exiting" )
                sys.exit(1)

        nname = node_in['__fdt_name__']
        nflag = LopperFDT.node_setname( fdt, nn, nname )
        if not nflag:
            lopper.log._error( f"node_sync: unable to set node {nn} name to {nname}" )
            sys.exit(1)

        try:
            ph = node_in['__fdt_phandle__']
            if ph:
                LopperFDT.property_set( fdt, nn, "phandle", ph )
        except:
            pass

        props = LopperFDT.node_properties( fdt, nn )
        props_to_delete = []
        for p in props:
            lopper.log._debug( f"node_sync: considering property {p.name} {p}" )
            if node_in['__fdt_phandle__'] and p.name == "phandle":
                # we just added this, it won't be in the node_in items under
                # the name name
                pass
            else:
                props_to_delete.append( p.name )

        lopper.log._debug( f"node_sync: properties pending deletion {props_to_delete}" )
        for prop, prop_val in reversed(node_in.items()):
            if re.search( r"^__", prop ) or prop.startswith( '/' ):
                lopper.log._debug( f"node_sync: skipping internal property {prop} ({prop_val})" )
                continue
            else:
                try:
                    qtype = node_in[f"__{prop}_type__" ]
                except:
                    qtype = None

                lopper.log._debug( f"node_sync: property {prop} value {prop_val} type {qtype}" )

                # We could supply a type hint via the __{}_type__ attribute
                LopperFDT.property_set( fdt, nn, prop, prop_val, LopperFmt.COMPOUND, verbose, qtype )
                if props_to_delete:
                    try:
                        props_to_delete.remove( prop )
                    except:
                        # if a node was added at the top of this routine, it
                        # won't have anything in the props_to_delete, and this
                        # would throw an exception. Since that's ok, we just
                        # catch it and move on.
                        pass

        for p in props_to_delete:
            lopper.log._debug( f"node_sync: deleting property {p}" )
            LopperFDT.property_remove( fdt, nname, p )

    @staticmethod
    def sync( fdt, dct, verbose = False ):
        """sync (write) a tree dictionary to a fdt

        This routine takes an input dictionary, and writes the details to
        the passed fdt.

        The dictionary contains a set of internal properties, as well as
        a list of standand properties to the node. Internal properties have
        a __ suffix and __ prefix.

        Child nodes are indexed by their absolute path. So any property that
        starts with "/" and is a dictionary, represents another node in the
        tree.

        In particular:
            - __path__ : is the absolute path fo the node, and is used to lookup
                         the target node
            - __fdt_name__ : is the name of the node and will be written to the
                             fdt name property
            - __fdt_phandle__ : is the phandle for the node

        All other non  '/' leading, or '__' leading properties will be written to
        the FDT as node properties.

        Passed nodes will be synced via the node_sync() function, and will
        be created if they don't exist. Existing nodes will have their properties
        deleted if they are not in the corresponding dictionary.

        All of the existing nodes in the FDT are read, if they aren not found
        in the passed dictionary, they will be deleted.

        Args:
            fdt (fdt): flattened device tree object
            node_in: (dictionary): Node description dictionary
            parent (dictionary,optional): parent node description dictionary
            verbose (bool,optional): verbosity level

        Returns:
            Nothing
        """
        # import a dictionary to a FDT

        if not fdt:
            return

        lopper.log._debug( "sync: start" )

        # we have a list of: containing dict, value, parent
        dwalk = [ [dct,dct,None]  ]
        node_ordered_list = []
        node_special_list = []
        while dwalk:
            firstitem = dwalk.pop()
            if type(firstitem[1]) is OrderedDict:
                node_ordered_list.append( [firstitem[1], firstitem[0]] )
                for item,value in reversed(firstitem[1].items()):
                    dwalk.append([firstitem[1],value,firstitem[0]])
            elif type(firstitem[1]) is dict:
                # type 'dict' are special nodes, that aren't order dependent,
                # gather them up for future processing.
                node_special_list.append( firstitem[1] )
            else:
                pass

        # this gets us a list of absolute paths. If we walk through the
        # dictionary passed in, and delete them from the list, we have the list
        # of nodes to delete with whatever is left over, and the nodes to add if
        # they aren't in the list.
        nodes_to_remove = LopperFDT.nodes( fdt, "/" )
        nodes_to_add = []
        for n_item in node_ordered_list:
            try:
                nodes_to_remove.remove( n_item[0]['__path__'] )
            except:
                nodes_to_add.append( n_item )

        for node in nodes_to_remove:
            nn = LopperFDT.node_find( fdt, node )
            if nn != -1:
                lopper.log._debug( f"sync: removing node {node}" )
                LopperFDT.node_remove( fdt, nn )
            else:
                lopper.log._debug( f"sync: node {node} not found; nothing to remove" )
                # child nodes are removed with their parent, and follow in the
                # list, so this isn't an error.
                pass

        for n in node_special_list:
            lopper.log._debug( f"sync: special node {n}" )
            if n['__path__'] == "/memreserve":
                memreserve_vals = n['__memreserve__']
                lopper.log._debug( f"sync: memreserve {memreserve_vals}" )
                # no swig wrapper, so we do this the hard way
                libfdt.fdt_add_mem_rsv( fdt._fdt, memreserve_vals[0], memreserve_vals[1] )

        # add the nodes
        for n in reversed(node_ordered_list):
            nn = LopperFDT.node_find( fdt, n[0]['__path__'] )
            if nn == -1:
                new_number = LopperFDT.node_add( fdt, n[0]['__path__'], True, verbose )
                if new_number == -1:
                    lopper.log._error( f"sync: node {n[0]['__path__']} could not be added; exiting" )
                    sys.exit(1)
                else:
                    lopper.log._debug( f"sync: node {n[0]['__path__']} added as offset {new_number}" )


        # were there any __lopper<>__ nodes created ? These were
        # temporary to work around libfdt prefix issues

        # sync the properties
        for n_item in reversed(node_ordered_list):
            node_in = n_item[0]
            node_in_parent = n_item[1]
            node_path = node_in['__path__']
            abs_path = node_path
            nn =  node_in['__fdt_number__']

            LopperFDT.node_sync( fdt, node_in, node_in_parent, verbose )

    @staticmethod
    def export( fdt, start_node = "/", verbose = False, strict = False, schema = None ):
        """export a FDT to a description / nested dictionary

        This routine takes a FDT, a start node, and produces a nested dictionary
        that describes the nodes and properties in the tree.

        The dictionary contains a set of internal properties, as well as
        a list of standand properties to the node. Internal properties have
        a __ suffix and __ prefix.

        Child nodes are indexed by their absolute path. So any property that
        starts with "/" and is a dictionary, represents another node in the
        tree.

        In particular:
            - __path__ : is the absolute path fo the node, and is used to lookup
                         the target node
            - __fdt_name__ : is the name of the node and will be written to the
                             fdt name property
            - __fdt_phandle__ : is the phandle for the node

        All other "standard" properties are returned as entries in the dictionary.

        if strict is enabled, structural issues in the input tree will be
        flagged and an error triggered. Currently, this is duplicate nodes, but
        may be extended in the future

        Args:
            fdt (fdt): flattened device tree object
            start_node (string,optional): the starting node
            verbose (bool,optional): verbosity level
            strict (bool,optional): toggle validity checking

        Returns:
            OrderedDict describing the tree
        """
        # export a FDT as a dictionary
        dct = OrderedDict()

        nodes = LopperFDT.node_subnodes( fdt, start_node )

        if strict:
            if len(nodes) != len(set(nodes)):
                raise Exception( f"lopper.fdt: duplicate node detected ({nodes})" )

        dct["__path__"] = start_node

        np = LopperFDT.node_properties_as_dict( fdt, start_node, schema=schema )
        if np:
            dct.update(np)

        nn = LopperFDT.node_number( fdt, start_node )
        dct["__fdt_number__"] = nn
        dct["__fdt_name__"] = LopperFDT.node_getname( fdt, start_node )
        dct["__fdt_phandle__"] = LopperFDT.node_getphandle( fdt, nn )

        lopper.log._debug( "export: start" )
        lopper.log._debug( f"export: startnode {start_node} subnodes {nodes}" )
        lopper.log._debug( f"export: properties {np}" )

        for i,n in enumerate(nodes):
            # Children are indexed by their path (/foo/bar), since properties
            # cannot start with '/'
            dct[n] = LopperFDT.export( fdt, n, verbose, strict, schema=schema )

        # only when processing the root node, we look to see if there
        # was a peer /memreserve node. if found, we add it to the exported
        # dictionary
        if start_node == "/":
            memreserve = fdt.num_mem_rsv()
            mdct = {}
            if memreserve:
                mdct["__fdt_number__"] = -1
                mdct["__fdt_name__"] = "memreserve"
                mdct["__fdt_phandle__"] = -1
                mdct["__path__"] = "/memreserve"

                lopper.log._debug( f"export: memreserve entries {memreserve}" )
                for idx in range(0,memreserve):
                    mr = fdt.get_mem_rsv(idx)
                    lopper.log._debug( f"export: memreserve {mr}" )

                    if type(mr) == int:
                        # we got a single number, which means our start address
                        # was 0x0, and the value is the range/size
                        mr = [ 0x0, mr ]

                    mdct["__memreserve__"] = mr

                dct["/memreserve"] = mdct


        return dct

    @staticmethod
    def node_properties_as_dict( fdt, node, type_hints=True, verbose=0, schema=None ):
        """Create a dictionary populated with the nodes properties.

        Builds a dictionary that is propulated with a node's properties as
        the keys, and their values. Used as a utility routine to avoid
        multiple calls to check if a property exists, and then to fetch its
        value.

        Args:
            fdt (fdt): flattened device tree object
            node (int or string): either a node number or node path
            type_hints  (bool,optional): flag indicating if type hints should be returned
            verbose (int,optional): verbosity level. default is 0.

        Returns:
            dict: dictionary of the properties, if successfull, otherwise and empty dict
        """
        prop_dict = {}

        # is the node a number ? or do we need to look it up ?
        node_number = -1
        node_path = ""
        try:
            node_number = int(node)
            node_path = LopperFDT.node_abspath( fdt, node )
        except ValueError:
            node_number = LopperFDT.node_find( fdt, node )
            node_path = node

        if node_number == -1:
            lopper.log._warning( f"could not find node {node_path}" )
            return prop_dict

        resolver = None
        if schema:
            resolver = lopper.schema.get_schema_manager().get_resolver()

        prop_list = LopperFDT.node_properties( fdt, node_path )
        for p in prop_list:
            if resolver:
                fmt_type = resolver.get_property_type(p.name, node_path)

                if p.name in lopper.schema.PROPERTY_DEBUG_SET:
                    lopper.log._debug( f"node_properties_as_dict: {node_path} {p.name}: schema type {fmt_type}" )

                if fmt_type != LopperFmt.UNKNOWN:
                    dtype = fmt_type
                else:
                    dtype = LopperFDT.property_type_guess( p )
            else:
                # this is the default of the function, but since we are always
                # passing an encoding, we need to explicitly assign this for the
                # no schema case to work. This triggers property type heuristics
                # at the lowest level
                fmt_type = LopperFmt.UNKNOWN
                dtype = LopperFDT.property_type_guess( p )

            property_val = LopperFDT.property_get( fdt, node_number, p.name,
                                                   LopperFmt.COMPOUND,
                                                   encode=fmt_type,
                                                   schema=schema )

            lopper.log._debug( f"node_properties_as_dict: fetched property with hint was: {property_val}")

            prop_dict[p.name] = property_val
            if type_hints:
                if dtype:
                    prop_dict[f'__{p.name}_type__'] = dtype

        lopper.log._debug( f"node {node_path} props as dict: {prop_dict}\n")

        return prop_dict

    @staticmethod
    def node_copy_from_path( fdt_source, node_source_path, fdt_dest, node_full_dest, verbose=0 ):
        """Copies a node from one FDT to another

        Copies a node between flattened device trees. The node (and
        properties) will be copied to the specified target device tree and
        path (ensure that a node does not already exist at the destination
        path).

        This routine is a wrapper around node_copy(), and will create a
        parent node structure in the destination fdt as required.

        Args:
            fdt_source (fdt): source flattened device tree object
            node_source_path: source device tree node path (fully specified)
            fdt_dest (fdt): destination flattened device tree object
            node_full_dest: destination device tree path for copied node (fully specified)
            verbose (int,optional): verbosity level. default is 0.

        Returns:
            bool: True if the node was copied, otherise, False
        """

        lopper.log._debug( f"node_copy_from_path: {node_source_path} -> {node_full_dest}" )

        node_to_copy = LopperFDT.node_find( fdt_source, node_source_path )
        node_dest_path = os.path.dirname( node_full_dest )
        node_dest_name = os.path.basename( node_full_dest )

        if node_dest_path == "/":
            node_dest_parent_offset = 0
        else:
            # non root dest
            node_dest_parent_offset = LopperFDT.node_find( fdt_dest, node_dest_path )
            if node_dest_parent_offset == -1:
                node_dest_parent_offset = LopperFDT.node_add( fdt_dest, node_dest_path )
                if node_dest_parent_offset <= 0:
                    lopper.log._error( "node_copy_from_path: unable to create destination node" )
                    sys.exit(1)

        if node_to_copy:
            return LopperFDT.node_copy( fdt_source, node_to_copy, fdt_dest, node_dest_parent_offset, verbose )

        return False

    @staticmethod
    def node_copy( fdt_source, node_source_offset, fdt_dest, node_dest_parent_offset, verbose=0 ):
        """Copies a node from one FDT to another

        Copies a node between flattened device trees. The node (and
        properties) will be copied to the specified target device tree and
        path (ensure that a node does not already exist at the destination
        path).

        Note: the destination node parent must exist before calling this routine

        Properties are iterated, decoded and then copied (encoded) to the
        destination node. As such, the copies are limited by the
        decode/encode capabilities. If properties do not look correct in the
        copy, the decode/encode routines need to be checked.

        Args:
            fdt_source (fdt): source flattened device tree object
            node_source_offset: source device tree node offset
            fdt_dest (fdt): destination flattened device tree object
            node_dest_parent_offset: destination device parent node
            verbose (int,optional): verbosity level. default is 0.

        Returns:
            bool: True if the node was copied, otherise, False
        """

        old_depth = -1
        depth = 0
        nn = node_source_offset
        newoff = node_dest_parent_offset
        while depth >= 0:
            nn_name = fdt_source.get_name(nn)
            for _ in range(MAX_RETRIES):
                try:
                    copy_added_node_offset = fdt_dest.add_subnode( newoff, nn_name )
                except Exception as e:
                    fdt_dest.resize( fdt_dest.totalsize() + 1024 )
                    continue
                else:
                    break
            else:
                lopper.log._error( "node_copy: unable to create destination subnode" )
                sys.exit(1)

            prop_offset = fdt_dest.subnode_offset( newoff, nn_name )

            lopper.log._debug( "", level=lopper.log.TRACE )
            lopper.log._debug( f"node_copy: properties for {fdt_source.get_name(nn)}", level=lopper.log.TRACE )

            # TODO: Investigate whether or not we can just copy the properties
            #       byte array directly. Versus decode -> encode, which could
            #       introduce issues.
            prop_list = []
            poffset = fdt_source.first_property_offset(nn, QUIET_NOTFOUND)
            while poffset > 0:
                prop = fdt_source.get_property_by_offset(poffset)

                # we insert, not append. So we can flip the order of way we are
                # discovering the properties
                prop_list.insert( 0, [ poffset, prop ] )

                lopper.log._debug( f"            prop name: {prop.name}", level=lopper.log.TRACE )
                lopper.log._debug( f"            prop raw: {prop}", level=lopper.log.TRACE )
                if lopper.log._is_enabled(lopper.log.TRACE):
                    prop_val = LopperFDT.property_value_decode( prop, 0 )
                    if not prop_val:
                        prop_val = LopperFDT.property_value_decode( prop, 0, LopperFmt.COMPOUND )
                    lopper.log._debug( f"            prop decoded: {prop_val}", level=lopper.log.TRACE )
                    lopper.log._debug( f"            prop type: {type(prop_val)}", level=lopper.log.TRACE )
                    lopper.log._debug( "", level=lopper.log.TRACE )

                poffset = fdt_source.next_property_offset(poffset, QUIET_NOTFOUND)

            # loop through the gathered properties and copy them over. We are reversing
            # the order of the way we iterated them, due to the way that setprop inserts
            # at zero every time. If we don't flip the order the copied node will have
            # them in the opposite order!

            for poffset, prop in prop_list:
                prop_val = LopperFDT.property_get( fdt_source, nn, prop.name, LopperFmt.COMPOUND )
                LopperFDT.property_set( fdt_dest, prop_offset, prop.name, prop_val )


            old_depth = depth
            nn, depth = fdt_source.next_node(nn, depth, (libfdt.BADOFFSET,))

            # we need a new offset fo the next time through this loop (but only if our depth
            # changed)
            if depth >= 0 and old_depth != depth:
                newoff = fdt_dest.subnode_offset( newoff, nn_name )

        return True

    @staticmethod
    def node_abspath( fdt, nodeid ):
        """Get the absolute (fully specified) path of a nodes

        Args:
            fdt (fdt): flattened device tree object
            nodeid: device tree node offset

        Returns:
            string: node path, if successful, otherwise ""
        """

        if nodeid == 0:
            return "/"

        node_id_list = [nodeid]
        p = fdt.parent_offset(nodeid,QUIET_NOTFOUND)
        while p != 0:
            node_id_list.insert( 0, p )
            p = fdt.parent_offset(p,QUIET_NOTFOUND)

        retname = ""
        for id in node_id_list:
            retname = retname + "/" + fdt.get_name( id )

        return retname

    @staticmethod
    def node_number( fdt, node ):
        """Get the number for the passed node

        Return the node number of a node by its path, or just return
        its number if it is already a number. This is a normalization
        routine for node references

        Args:
            fdt (fdt): flattened device tree object
            node (string or ing): the name or node number to check

        Returns:
            string: node number, or -1 if the node doesn't exist
        """
        # is the node a number ? or do we need to look it up ?
        node_number = -1

        try:
            node_number = int(node)
        except ValueError:
            node_number = LopperFDT.node_find( fdt, node )
            if node_number == -1:
                node_number, matching_nodes = LopperFDT.node_find_by_name( fdt, node )
            else:
                # libfdt returns matches based on partial paths. i.e. if a tree
                # has /timer@2a810000 and /timer, and you search for /timer, you
                # will get the node number for /timer@2a810000 .. this causes problems.
                # so we double check the name, and if it isn't as expected, we do a
                # slower secondary search to find the proper node.
                node_name = fdt.get_name( node_number )
                if node_name != os.path.basename( node):
                    node_number, matching_nodes = LopperFDT.node_find_by_name( fdt, os.path.basename(node) + "$" )

        if node_number == -1:
            lopper.log._warning( f"could not find node {node}" )

        return node_number

    @staticmethod
    def nodes_with_property( fdt, match_propname, match_regex="",
                             start_path="/", include_children=True, match_depth=0 ):
        """Get a list of nodes with a particular property

        Searches a device tree and returns a list of nodes that contain
        a given property.

        Matching is done by the existence of a property name in a node.

        If a match_regex is passed, then the value of the property is
        tested against the regex. If there's a match, then the node is
        added to the list.

        Args:
            fdt (fdt): source flattened device tree to search
            match_propname (string): target property name
            match_regex (string,optional): property value match regex. Default is ""
            start_path (string,optional): starting path in the device tree. Default is "/"
            include_children (bool,optional): should child nodes be searched. Default is True.
            match_depth (int): depth of the node, relative to the start path. Default is 0 (all nodes)

        Returns:
            list: list of matching nodes if successful, otherwise an empty list
        """

        # node_list = []
        depth = 0
        ret_nodes = []
        if start_path != "/":
            node, nodes = LopperFDT.node_find_by_name( fdt, start_path )
        else:
            node = 0

        if node < 0:
            lopper.log._error( f"could not find starting node {start_path}" )
            sys.exit(1)

        while depth >= 0:
            prop_val = LopperFDT.property_get( fdt, node, match_propname, LopperFmt.COMPOUND )
            if match_depth > 0:
                if match_depth != depth:
                    prop_val = None

            if prop_val:
                if match_regex:
                    for p in prop_val:
                        if re.search( match_regex, p ):
                            ret_nodes.append(node)
                else:
                    if match_propname == prop.name:
                        if not node in ret_nodes:
                            ret_nodes.append(node)

            node, depth = fdt.next_node(node, depth, (libfdt.BADOFFSET,))

        return ret_nodes

    @staticmethod
    def write_fdt( fdt_to_write, output_filename, overwrite=True, verbose=0, enhanced=False ):
        """Write a system device tree to a file

        Write a fdt (or system device tree) to an output file. This routine uses
        the output filename to determine if a module should be used to write the
        output.

        If the output format is .dts or .dtb, Lopper takes care of writing the
        output. If it is an unrecognized output type, the available assist
        modules are queried for compatibility. If there is a compatible assist,
        it is called to write the file, otherwise, a warning or error is raised.

        Args:
            fdt_to_write (fdt): source flattened device tree to write
            output_filename (string): name of the output file to create
            overwrite (bool,optional): Should existing files be overwritten. Default is True.
            verbose (int,optional): verbosity level to use.
            enhanced(bool,optional): whether enhanced printing should be performed. Default is False

        Returns:
            Nothing

        """
        if not output_filename:
            return

        if re.search( r"\.dtb$", output_filename ):
            lopper.log._info( f"write_fdt: dtb output format detected; writing {output_filename}" )

            byte_array = fdt_to_write.as_bytearray()

            lopper.log._info( f"write_fdt: writing output dtb {output_filename}" )

            o = Path(output_filename)
            if o.exists() and not overwrite:
                lopper.log._error( f"write_fdt: output file {output_filename} exists and overwrite disabled" )
                sys.exit(1)

            with open(output_filename, 'wb') as w:
                w.write(byte_array)

        elif re.search( r"\.dts$", output_filename ):
            lopper.log._info( f"write_fdt: dts output format detected; writing {output_filename}" )

            o = Path(output_filename)
            if o.exists() and not overwrite:
                lopper.log._error( f"write_fdt: output file {output_filename} exists and overwrite disabled" )
                sys.exit(1)

            if enhanced:
                printer = LopperTreePrinter( True, output_filename, verbose )
                printer.load( LopperFDT.export( fdt_to_write ) )
                printer.exec()
            else:
                # write the device tree to a temporary dtb
                fp = tempfile.NamedTemporaryFile()
                byte_array = fdt_to_write.as_bytearray()
                with open(fp.name, 'wb') as w:
                    w.write(byte_array)

                LopperFDT.dtb_dts_export( fp.name, output_filename )

                # close the temp file so it is removed
                fp.close()
        else:
            lopper.log._info( f"write_fdt: unknown file type ({output_filename}); skipping" )

    @staticmethod
    def dtb_dts_export( dtb, outfilename="", verbose=0 ):
        """writes a dtb to a file or to stdout as a dts

        Args:
           dtb: a compiled device tree
           outfilename (string): the output filename (stdout is used if empty)
           verbose (int,optional): extra debug info. default 0.

        Returns:
           The return value of executing dtc to dump the dtb to dts
        """
        dtcargs = (os.environ.get('LOPPER_DTC') or shutil.which("dtc")).split()
        dtcargs += (os.environ.get("LOPPER_DTC_FLAGS") or "").split()
        dtcargs += (os.environ.get("LOPPER_DTC_BFLAGS") or "").split()
        if outfilename:
            dtcargs += ["-o", f"{outfilename}"]
        dtcargs += ["-I", "dtb", "-O", "dts", dtb]

        lopper.log._info( f"dtb_dts_export: dumping dtb with args {dtcargs}" )

        result = subprocess.run(dtcargs, check = False, stderr=subprocess.PIPE )
        if result.returncode != 0:
            lopper.log._error( "dtb_dts_export: unable to export dts" )
            lopper.log._error( f"\n{textwrap.indent(result.stderr.decode(), '         ')}" )

        return result

    @staticmethod
    def dt_to_fdt( dtb, rmode='rb' ):
        """takes a dtb and returns a flattened device tree object

        Args:
           dtb: a compiled device tree
           rmode (string,optional): the read mode of the file, see libfdt for possible values
                                    default is 'rb'

        Returns:
           A flattended device tree object (as defined by libfdt)
        """
        fdt = libfdt.Fdt(open(dtb, mode=rmode).read())
        return fdt

    @staticmethod
    def node_getphandle( fdt, node_number ):
        """utility command to get a phandle (as a number) from a node

        Args:
           fdt (FDT): flattened device tree
           node_number (int): node number in the fdt

        Returns:
           int: the phandle of the node number, if successful, -1 if not
        """
        prop = fdt.get_phandle( node_number )
        return prop


    @staticmethod
    def property_get( fdt, node_number, prop_name, ftype=LopperFmt.SIMPLE,
                      encode=LopperFmt.DEC, schema=None ):
        """utility command to get a property (as a string) from a node

        A more robust way to get the value of a property in a node, when
        you aren't sure of the format of that property. This routine takes
        hints when getting the property in the form of a "format type" and
        an encoding.

        The format and encoding options are in the following enum type:

           class LopperFmt(Enum):
              SIMPLE = 1 (format)
              COMPOUND = 2 (format)
              HEX = 3 (encoding)
              DEC = 4 (encoding)
              STRING = 5 (encoding)
              MULTI_STRING = 5 (encoding)

        Args:
           fdt (FDT): flattened device tree
           node_number (int): node number in the fdt
           property (string): property name whose value to get
           ftype (LopperFmt,optional): format of the property. Default SIMPLE.
           encode (LopperFmt,optional); encoding of the property. Default DEC

        Returns:
           string: if format is SIMPLE: string value of the property, or "" if not found
           list: if format is COMPOUND: list of property values as strings, [] if not found
        """
        try:
            lopper.log._debug( f"property_get: {prop_name}, encoding is: {encode} ftype is {ftype}" )
            prop = fdt.getprop( node_number, prop_name )
            val = LopperFDT.property_value_decode( prop, 0, ftype, encode )
            lopper.log._debug( f"property_get: decoded as {val}" )
        except Exception as e:
            val = ""

        return val

    @staticmethod
    def property_set( fdt, node_number, prop_name, prop_val, ftype=LopperFmt.SIMPLE,
                      verbose=False, typehint=None ):
        """utility command to set a property in a node

        A more robust way to set the value of a property in a node, This routine
        takes hints when getting the property in the form of a "format type"

        The format options are in the following enum type:

           class LopperFmt(Enum):
              SIMPLE = 1 (format)
              COMPOUND = 2 (format)

        Based on the format hint, and the passed value, the property is encoded
        into a byte array and stored into the flattened device tree node.

        Args:
           fdt_dst (FDT): flattened device tree
           node_number (int): node number in the fdt
           prop_name (string): property name whose value to set
           ftype (LopperFmt,optional): format of the property. Default SIMPLE.

        Returns:
           Nothing

        """

        # if it's a list, we dig in a bit to see if it is a single item list.
        # if so, we grab the value so it can be propery encoded. We also have
        # a special case if the '' string is the only element .. we explicity
        # set the empty list, so it will encode properly.
        if type(prop_val) == list:
            if len(prop_val) == 1 and prop_val[0] != '':
                prop_val = prop_val[0]
            elif len(prop_val) == 1 and prop_val[0] == '':
                pass

        try:
            prop_val_converted = int(prop_val,0)
            # if it works, that's our new prop_val. This covers the case where
            # a string is passed in, but it is really just a single number.
            # note: we may need to consult "ftype" in the future so the caller
            # can override this automatical conversion
            prop_val = prop_val_converted
        except:
            # do nothing. let propval go through as whatever it was
            pass

        # we have to re-encode based on the type of what we just decoded.
        if type(prop_val) == int or type(prop_val) == bool:
            # this seems to break some operations, but a variant may be required
            # to prevent overflow situations
            # if sys.getsizeof(prop_val) >= 32:
            for _ in range(MAX_RETRIES):
                try:
                    if typehint and typehint == LopperFmt.UINT8:
                        bval = LopperFDT.encode_byte_array([prop_val], 1)
                        fdt.setprop( node_number, prop_name, bval)
                    elif sys.getsizeof(prop_val) > 32:
                        fdt.setprop_u64( node_number, prop_name, prop_val )
                    else:
                        fdt.setprop_u32( node_number, prop_name, prop_val )

                    break
                except Exception as e:
                    fdt.resize( fdt.totalsize() + 1024 )
                    continue
                else:
                    break
            else:
                # it wasn't set all all, we could thrown an error
                pass
        elif type(prop_val) == str:
            for _ in range(MAX_RETRIES):
                try:
                    fdt.setprop_str( node_number, prop_name, prop_val )
                    break
                except Exception as e:
                    lopper.log._debug( f"property_set: exception {e}" )
                    fdt.resize( fdt.totalsize() + 1024 )
                    continue
                else:
                    break
            else:
                # we totally failed!
                pass

        elif type(prop_val) == list:

            if len(prop_val) > 1:
                val_to_sync = []
                iseq = iter(prop_val)
                first_type = type(next(iseq))
                # check for a mixed type, we get "false" if it is not all the same, or
                # the type otherwise
                the_same = first_type if all( (type(x) is first_type) for x in iseq ) else False
                if the_same == False:
                    # convert everything to strings
                    val_to_sync = []
                    for v in prop_val:
                        val_to_sync.append( str(v) )
                else:
                    val_to_sync = prop_val
            else:
                val_to_sync = prop_val

            prop_val = val_to_sync

            # list is a compound value, or an empty one!
            if len(prop_val) >= 0:
                if typehint and typehint == LopperFmt.UINT8:
                    bval = LopperFDT.encode_byte_array(prop_val, 1)
                else:
                    try:
                        bval = LopperFDT.encode_byte_array_from_strings(prop_val)
                    except Exception as e:
                        bval = LopperFDT.encode_byte_array(prop_val)

                for _ in range(MAX_RETRIES):
                    try:
                        fdt.setprop( node_number, prop_name, bval)
                    except Exception as e:
                        fdt.resize( fdt.totalsize() + 1024 )
                        continue
                    else:
                        break
                else:
                    # fail!
                    lopper.log._warning( f"property_set: unable to write property '{prop_name}' to fdt" )
        else:
            lopper.log._warning( f"property_set: {prop_name} has unsupported type {type(prop_val)}" )

    @staticmethod
    def property_remove( fdt, node_name, prop_name, verbose=0 ):
        """removes a property from a fdt

        Removes a property (if it exists) from a node (and optionally its children).

        Args:
            fdt (FDT): flattened device tree to modify
            node_name (int or string): the node number or name to process
            prop_name (string): name of property to remove

        Returns:
            Boolean: True if the property was deleted, False if it wasn't

        """

        node = LopperFDT.node_find( fdt, node_name )
        if node == -1:
            node, nodes = LopperFDT.node_find_by_name( fdt, node_name )

        if node == -1:
            return False

        prop_list = []
        poffset = fdt.first_property_offset(node, QUIET_NOTFOUND)
        while poffset > 0:
            # if we delete the only property of a node, all calls to the FDT
            # will throw an except. So if we get an exception, we set our poffset
            # to zero to escape the loop.
            try:
                prop = fdt.get_property_by_offset(poffset)
            except:
                poffset = 0
                continue

            prop_list.append(prop.name)
            poffset = fdt.next_property_offset(poffset, QUIET_NOTFOUND)

        if prop_name in prop_list:
            # node is an integer offset, prop_name is a string
            lopper.log._info( f"property_remove: removing {prop_name} from {fdt.get_name(node)}" )

            fdt.delprop(node, prop_name)
        else:
            return False

        return True

    @staticmethod
    def dt_compile( dts_file, i_files, includes, force_overwrite=False, outdir="./",
                    save_temps=False, verbose=0, enhanced = True, permissive = False,
                    symbols=False):
        """Compile a dts file to a dtb

        This routine takes a dts input file, other dts include files,
        include search path and then uses standard tools (cpp, dtc, etc).

        Environment variables can be used tweak the execution of the various
        tools and stages:

           LOPPER_CPP: set if a different cpp than the standard one should
                       be used, or if cpp is not on the path
           LOPPER_PPFLAGS: flags to be used when calling cpp
           LOPPER_DTC: set if a non standard dtc should be used, or if dtc
                       is not on the path
           LOPPER_DTC_FLAGS: flags to use when calling dtc
           LOPPER_DTC_OFLAGS: extra dtc flags if an overlay is being compiled
           LOPPER_DTC_BFLAGS: extra dtc args/flags

        Args:
           dts_file (string): path to the dts file to be compiled
           i_files (list): files to be included
           includes (list): list of include directories (translated into -i <foo>
                            for dtc calls)
           force_overwrite (bool,optional): should files be overwritten.
                                            Default is False
           save_temps (bool, optional): should temporary files be saved on failure
           symbols (bool,optional) : should __symbols__ node be created
           verbose (bool,optional): verbosity level

        Returns:
           string: Name of the compiled dtb

        """
        output_dtb = ""

        # Note: i_files is not currently used. They are typically concatenated
        #       before calling this routine due to pecularities in include
        #       processing
        # TODO: might need to make 'dts_file' absolute for the cpp call below
        dts_filename = os.path.basename( dts_file )
        dts_filename_noext = os.path.splitext(dts_filename)[0]

        #
        # step 1: preprocess the file with CPP (if available)
        #

        # NOTE: we are putting the .pp file into the same directory as the
        #       system device tree. Without doing this, dtc cannot resolve
        #       labels from include files, and will throw an error. If we get
        #       into a mode where the system device tree's directory is not
        #       writeable, then we'll have to either copy everything or look
        #       into why dtc can't handle the split directories and include
        #       files.

        preprocessed_name = LopperFDT.dt_preprocess( dts_file, includes, outdir, verbose )

        # even without the enhanced flag we parse phandle information from
        # the dts file. This allows us to do analysis and error checking
        # later on new properties that may have phandles.
        with open( preprocessed_name, 'r') as file:
            pdata = file.read()

        generator = lopper.schema.DTSSchemaGenerator()
        phandle_descriptions, phandle_map = generator.scan_dts_file(pdata)
        schema = generator.generate_schema()

        # This (the phandle_map saving) can become optional and used
        # only when a command line flag is passed. The map is quite
        # large and is a slower lookup.
        phandle_dts = lopper_base.encode_phandle_map_to_dts( phandle_map )

        if enhanced:
            fp = preprocessed_name

            # we need to ensure comments are maintained by converting them
            # into DTS attributes
            fp_enhanced = fp + ".enhanced"
            shutil.copyfile( fp, fp_enhanced )
            fp = fp_enhanced

            with open(fp_enhanced, 'r') as file:
                data = file.read()

            # drop /dts-v1/; from the file, we'll add it back at the top first. but
            # for now, we need it out of te way to look for any preamble to the main
            # device tree nodes

            dts_regex = re.compile( r'\/dts-v1/;' )
            if re.search( dts_regex, data ):
                # delete the dts opening, since we are going to capture everything
                # from the start of the file, to the opening of the device tree
                # nodes.
                data = re.sub( dts_regex, '', data )

            memres_string = ""
            memres_regex = re.compile( r'\/memreserve\/(.*)?;' )
            memres = re.search( memres_regex, data )
            if memres:
                lopper.log._debug( f"memreserve detected: {memres.group(1)}" )
                memres_string = f"/memreserve/ {memres.group(1)};"

            # This captures everything at the start of the file (i.e. a comment block)
            # and puts it into a special pre-mble property in the root node. If we don't
            # do this, and let the comment substituion find it below, we have an invalid
            # device tree.
            #
            # When printing the tree layer, we'll pop it out and put it back as an
            # opening comment.
            #
            preamble_regex = re.compile( r'(^.*?)(/ {)', re.MULTILINE | re.DOTALL )
            preamble = re.search( preamble_regex, data )
            if preamble:
                # is it a comment block ? if so, we want to mark it specially so
                # it can be put back at the header later.
                comment_regex = re.compile( r'(/\*)(.*?)(\*/)', re.MULTILINE | re.DOTALL )
                comment = re.search( comment_regex, preamble.group(1) )
                if comment:
                    comment = comment.group(2)
                    if comment:
                        comment = re.sub( r"^\n", '', comment )
                        comment = re.sub( r"\n$", '', comment )
                        comment = f"    lopper-preamble = \"{comment}\";"

                    data = re.sub( preamble_regex, '/ {' + f'\n\n{comment}', data, count = 1 )

            # put the dts start info back in
            data = re.sub( r'^', f'/dts-v1/;\n\n{memres_string}\n', data )

            # Comment and label substitution
            fp_comments_as_attributes = LopperFDT._comment_translate(data)
            fp_comments_and_labels_as_attributes = LopperFDT._label_translate(fp_comments_as_attributes)

            # now we need to potentially sort/drop some comments that are in bad places
            #   - comments must preceed nodes
            #   - comments cannot be in the middle of a sequence
            #
            # In the future we could always try and relocate them, but we've done our
            # best at this pont, so we'll just delete them

            lopper_comment_pattern = re.compile(r'lopper-comment-([0-9]+) = "(.*?)"', re.DOTALL | re.MULTILINE )
            lopper_comment_pattern_with_context = re.compile(r'{(.*?)lopper-comment-([0-9]+) = "(.*?)"', re.DOTALL | re.MULTILINE )
            lopper_comment_open_pattern = re.compile(r'lopper-comment-([0-9]+) = "(.*$)', re.DOTALL )

            comments_to_delete = []
            file_as_array = fp_comments_and_labels_as_attributes.splitlines()
            file_boundary_index = 0
            node_depth = 0
            subnode_at_depth = { 0: False }
            for i,f in enumerate(file_as_array):
                # take note when we pass an included file boundary, we'll use it to scan back
                # and look for structures.
                ml = re.search( r"^\#line.*\"(.*?)\"", f )
                if ml:
                    lopper.log._debug( f"comment scan: include file boundary passed {ml.group(1)}" )
                    file_boundary_index = i
                    # clear the node tracking counts, we are into a new file
                    subnode_at_depth = { 0: False }
                    node_depth = 0

                mn = re.search( r"^\s*(.*){", f )
                if mn:
                    node_depth += 1
                    lopper.log._debug( f"comment scan: node depth increased to {node_depth} ({mn.group(1)})" )
                    subnode_at_depth[node_depth-1] = True
                    subnode_at_depth[node_depth] = False

                mn = re.search( r"^\s*};", f )
                if mn:
                    lopper.log._debug( f"comment scan: node depth decreased to {node_depth}" )
                    node_depth -= 1

                m = re.search( lopper_comment_open_pattern, f )
                if m:
                    comment_number = m.group(1)
                    lopper.log._debug( f"comment scan: line {i} has comment #{comment_number} [{m.group(2)}]" )

                    if subnode_at_depth[node_depth]:
                        lopper.log._debug( "comment scan: comment found after first subnode, tagging to delete" )
                        comments_to_delete.append( comment_number )

                    if node_depth == 0:
                        lopper.log._debug( "comment scan: comment before any nodes, tagging to delete" )
                        comments_to_delete.append( comment_number )

                    m2 = re.search( r"^\s*lopper-comment", file_as_array[i] )
                    if not m2:
                        m3 = re.search( r"[{;]\s*lopper-comment", file_as_array[i] )
                        if not m3:
                            lopper.log._debug( "comment scan: comment embedded in property, tagging to delete" )
                            comments_to_delete.append( comment_number )

            # in case our comment fixups miss something, pull an env variable that is a list
            # of comment numbers to delete.
            extra_drops = os.environ.get('LOPPER_COMMENT_DROPLIST')
            if extra_drops:
                comments_to_delete.extend( extra_drops.split() )
                lopper.log._debug( f"comment scan: droplist {comments_to_delete}" )
            for cnum in comments_to_delete:
                lopper_comment_regex = re.compile( r'lopper-comment-{0} = ".*?";'.format(cnum), re.MULTILINE | re.DOTALL )
                fp_comments_and_labels_as_attributes = re.sub( lopper_comment_regex, "", fp_comments_and_labels_as_attributes )

            labeldict = {}
            lopper_label_pattern = re.compile(r'lopper-label-([0-9]+) = "(.*?)"')
            tree_block_pattern = re.compile( r'/ {' )

            file_as_array = fp_comments_and_labels_as_attributes.splitlines()
            for i,f in enumerate(file_as_array):
                m = re.search( tree_block_pattern, f )
                if m:
                    labeldict = {}

                labelnum = 0
                label = None
                m = re.search( lopper_label_pattern, f )
                if m:
                    labelnum = m.group(1)
                    label = m.group(2)

                if label:
                    try:
                        existing_label = labeldict[label]
                        if not permissive:
                            lopper.log._error( f"duplicate label '{label}' detected; processing cannot continue" )
                            lopper.log._debug( "label processing: dumping label dictionary (as processed to error)" )
                            for l in labeldict:
                                lopper.log._debug( f"    {l}" )

                            lopper.log._debug( "\noffending label lines with context:" )
                            file_as_array = data.splitlines()
                            pattern = re.compile( r'^\s*?({})\s*?\:(.*?)$'.format(label), re.DOTALL | re.MULTILINE )
                            match_line = 0
                            for i,f in enumerate(file_as_array):
                                m = re.search( pattern, f )
                                if m:
                                    try:
                                        lopper.log._debug( f"    {i - 2} {file_as_array[i - 2]}" )
                                        lopper.log._debug( f"    {i - 1} {file_as_array[i - 1]}" )
                                        lopper.log._debug( f"    {i} {file_as_array[i]}" )
                                        lopper.log._debug( f"    {i + 1} {file_as_array[i + 1]}" )
                                        lopper.log._debug( f"    {i + 2} {file_as_array[i + 2]}" )
                                    except:
                                        lopper.log._debug( f"    {i} {file_as_array[i]}" )

                                    lopper.log._debug( "\n" )
                            os._exit(1)
                        else:
                            # clobber the dup
                            labeldict[label] = label
                    except:
                        labeldict[label] = label

            f = open( fp_enhanced, 'w' )

            # add our phandle tracking node
            fp_comments_and_labels_as_attributes +=  "\n" + phandle_dts

            f.write( fp_comments_and_labels_as_attributes )
            f.close()

            preprocessed_name = fp_enhanced

        # step 2: compile the dtb
        #         dtc -O dtb -o test_tree1.dtb test_tree1.dts
        isoverlay = False
        output_dtb = f"{dts_filename}.{'dtbo' if isoverlay else 'dtb'}"

        # make sure the dtb is not on disk, since it won't be overwritten by
        # default.
        if os.path.exists( output_dtb ):
            if not force_overwrite:
                lopper.log._error( f"output dtb ({output_dtb}) exists and -f was not passed" )
                sys.exit(1)
            os.remove( output_dtb )

        dtcargs = (os.environ.get('LOPPER_DTC') or shutil.which("dtc")).split()
        dtcargs += (os.environ.get( 'LOPPER_DTC_FLAGS') or "").split()
        if isoverlay:
            dtcargs += (os.environ.get("LOPPER_DTC_OFLAGS") or "").split()
        else:
            dtcargs += (os.environ.get("LOPPER_DTC_BFLAGS") or "").split()
        for i in includes.split():
            dtcargs += ["-i", i]

        dtcargs += ["-o", f"{outdir}/{output_dtb}"]
        if symbols:
            dtcargs += [ "-@" ]
        dtcargs += ["-I", "dts", "-O", "dtb", preprocessed_name ]
        lopper.log._info( f"dt_compile: compiling dtb with args {dtcargs}" )

        result = subprocess.run(dtcargs, check = False, stderr=subprocess.PIPE )
        if result.returncode != 0:
            # force the dtb, we need to do processing
            dtcargs += [ "-f" ]
            lopper.log._info( f"dt_compile: forcing dtb generation with args {dtcargs}" )

            result = subprocess.run(dtcargs, check = False, stderr=subprocess.PIPE )
            if result.returncode != 0:
                lopper.log._error( f"dt_compile: unable to compile with args {dtcargs}" )
                lopper.log._error( f"\n{textwrap.indent(result.stderr.decode(), '         ')}" )
                sys.exit(1)

        # cleanup: remove the .pp file
        if not save_temps:
            os.remove( preprocessed_name )

        # if we got here, and for some reason the output_dtb does not exist, we should
        # zero the name and return "" instead.
        output_file = Path(outdir + "/" + output_dtb)
        try:
            output_file_path = output_file.resolve()
        except FileNotFoundError:
            output_dtb = ""

        return [ str(output_file), schema ]



# /*
# * Copyright (c) 2021 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Izhar Shaikh <izhar.ameer.shaikh@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

from collections import OrderedDict
from operator import ior
from functools import reduce

nodes = dict()
debug = False

# Constants
NS_CHECK_TYPE = False  # Region checking policy
TZ = True  # Set trustzone true for default masters
RD = 1  # read okay
WR = 1  # read/write okay

DEF_RD_ALLOWED = False
DEF_WR_ALLOWED = False

# Interrupts
Interrupts = {
    "IEN_SECURITY_VIO": [True, 3],  # Enable Security violation interrupt
    "IEN_WRPERM_VIO": [True, 2],  # Enable Write permission violation interrupt
    "IEN_RDPERM_VIO": [True, 1],  # Enable Read permission Violation interrupt
    "IEN_INV_APB": [True, 0],  # Enable Register Access interrupt on APB
}

# Total regions
REGIONS = 16  # 16 Regions

# Aper Address start
REGION_0_START = 0x100
REGION_STEP_SIZE = 0x18

# REGION offsets
REGN_X_START_LO_OFFSET = 0x0
REGN_X_START_HI_OFFSET = 0x4
REGN_X_END_LO_OFFSET = 0x8
REGN_X_END_HI_OFFSET = 0xC
REGN_X_MASTER_OFFSET = 0x10
REGN_X_CONFIG_OFFSET = 0x14

# Interrupt
IEN_OFFSET = 0x18

OFFSETS = {
    "s_lo": REGN_X_START_LO_OFFSET,
    "s_hi": REGN_X_START_HI_OFFSET,
    "e_lo": REGN_X_END_LO_OFFSET,
    "e_hi": REGN_X_END_HI_OFFSET,
    "master": REGN_X_MASTER_OFFSET,
    "config": REGN_X_CONFIG_OFFSET,
}

# Master ID List (Global)
MIDL = {
    "PPU0": [0x246, 0x3FF],
    "PPU1": [0x247, 0x3FF],
    "PPU0/PPU1": [0x246, 0x3FE],
    "DAP": [0x240, 0x3FF],
    "HSDP_DPC": [0x249, 0x3FF],
    "PMC_DMA0": [0x248, 0x3FF],
    "PMC_DMA1": [0x24B, 0x3FF],
    "APU": [0x260, 0x3F0],
    "APU0": [0x260, 0x3FF],
    "APU1": [0x261, 0x3FF],
    "RPU0": [0x200, 0x3FF],
    "RPU1": [0x204, 0x3FF],
    "PSM": [0x238, 0x3FF],
    # -- #
    "ANY": [0x0, 0x0],
}

# default masters list
DEF_MASTERS = [
    "PPU0", "PPU1", "PSM", "DAP", "HSDP_DPC", "PMC_DMA0", "PMC_DMA1"
]


def h2i(string):
    if string is None:
        return 0
    return int(string, 16)


def i2h(num):
    return hex(num)


def dword(num):
    return (num & 0xffffffff)


class Master:
    """XMPU MASTER_IDXX"""
    def __init__(self, mid, midm, name=''):
        self.name = name  # Master name string
        self.mid = mid  # Master Id (SMID)
        self.mask = midm  # Master Id Mask

    def __str__(self):
        return str(self.__class__) + ": " + str(self.__dict__)

    def val(self):
        midm = self.mask << 16
        mid = self.mid
        return midm + mid

    def val_s(self):
        return hex(self.val())


class Config:
    """ XMPU Region Config """
    def __init__(self, check_type, tz, wr_allowed, rd_allowed, enable):
        self.check_type = check_type
        self.tz = tz
        self.wr_allowed = wr_allowed
        self.rd_allowed = rd_allowed
        self.enable = enable

    def __str__(self):
        return str(self.__class__) + ": " + str(self.__dict__)

    def val(self):
        bits = [
            self.enable, self.rd_allowed, self.wr_allowed, self.tz,
            self.check_type
        ]
        bitv = (bit << pos for pos, bit in enumerate(bits))
        return reduce(ior, bitv)

    def set_tz(self, tz):
        self.tz = tz

    def set_wr_allowed(self, wr_allowed):
        self.wr_allowed = wr_allowed

    def set_rd_allowed(self, rd_allowed):
        self.rd_allowed = rd_allowed

    def set_rw(self, rw):
        if rw == 0:
            self.set_wr_allowed(1)
            self.set_rd_allowed(1)
        elif rw == 1:
            self.set_rd_allowed(1)
        elif rw == 2:
            self.set_wr_allowed(1)
        else:
            return

    def en(self):
        self.enable = 1

    def disable(self):
        self.enable = 0


class Region:
    """ XMPU Regions """
    def __init__(self, idx, offset):
        self.idx = idx
        self.offset = offset
        self.addr_start = 0
        self.addr_end = 0
        self.master = Master(0, 0)
        self.config = Config(NS_CHECK_TYPE, TZ, WR, RD, 0)

    def __str__(self):
        return str(self.__class__) + ": " + str(self.__dict__)

    def set_master(self, master: Master):
        self.master.mid = master.mid
        self.master.mask = master.mask
        self.master.name = master.name

    def set_addr(self, addr, size):
        self.addr_start = addr
        self.addr_end = addr + size

    def get_offset(self, offkey):
        return self.offset + OFFSETS[offkey]

    def get_master(self):
        return self.get_offset("master"), dword(self.master.val())

    def get_config(self):
        return self.get_offset("config"), dword(self.config.val())

    def get_start_lo(self):
        return self.get_offset("s_lo"), dword(self.addr_start)

    def get_start_hi(self):
        return self.get_offset("s_hi"), dword(self.addr_start >> 32)

    def get_end_lo(self):
        return self.get_offset("e_lo"), dword(self.addr_end)

    def get_end_hi(self):
        return self.get_offset("e_hi"), dword(self.addr_end >> 32)


class Xmpu:
    """Xmpu base class"""
    def __init__(self, name, addr, size):
        self.name = name
        self.baseaddr = addr
        self.size = size
        self.current = 0
        # initialize regions map
        self.regions = {
            i: Region(i, REGION_0_START + (i * REGION_STEP_SIZE))
            for i in range(REGIONS)
        }

    def __str__(self):
        return str(self.__class__) + ": " + str(self.__dict__)

    def set_master(self, idx, master: Master):
        if idx < len(self.regions):
            self.regions[idx].set_master(master)
        else:
            print("[ERROR] xmpu region idx {} too big (max: {})".format(
                idx, REGIONS))
            return

    def set_addr(self, idx, addr, size):
        if idx < len(self.regions):
            self.regions[idx].set_addr(addr, size)
        else:
            print("[ERROR] xmpu region idx {} too big (max: {})".format(
                idx, REGIONS))
            return

    def set_config(self, idx, tz, rw):
        if idx < len(self.regions):
            self.regions[idx].config.set_tz(tz)
            self.regions[idx].config.set_rw(rw)
        else:
            print("[ERROR] xmpu region idx {} too big (max: {})".format(
                idx, REGIONS))
            return

    def enable_region(self, idx):
        if idx < len(self.regions):
            self.regions[idx].config.en()
        else:
            print("[ERROR] xmpu region idx {} too big (max: {})".format(
                idx, REGIONS))
            return

    def create_region_and_en(self, addr, size, smid, mask, rw, tz, name=''):
        if self.current >= REGIONS:
            print("[ERROR] {} xmpu regions full!".format(self.name))
            return False

        self.set_addr(self.current, addr, size)
        self.set_master(self.current, Master(smid, mask, name=name))
        self.set_config(self.current, tz, rw)
        self.enable_region(self.current)
        self.current += 1
        return True

    def is_filled(self):
        return self.current < REGIONS

    def get_ctrl_reg_addr_val(self):
        return hex(self.baseaddr), hex((DEF_WR_ALLOWED << 1)
                                       | (DEF_RD_ALLOWED))

    def get_ien_reg_addr_val(self):
        reg_addr = self.baseaddr + IEN_OFFSET
        reg_val = 0
        for istr, ival in Interrupts.items():
            reg_val = reg_val | (ival[0] << ival[1])
        return hex(reg_addr), hex(reg_val)


def mid(master):
    if master in MIDL:
        mid = MIDL[master][0]
        midm = MIDL[master][1]
    else:
        print("Error: missing MID in the list!")
    return Master(mid, midm, name=master)


def init_masters(xmpu):
    xmpu.set_master(15, mid("PPU0/PPU1"))  # PPU0/PPU1 (PMC)
    xmpu.set_master(14, mid("PSM"))  # PSM
    xmpu.set_master(13, mid("PMC_DMA0"))  # PMC DMA0
    xmpu.set_master(12, mid("PMC_DMA1"))  # PMC DMA1
    xmpu.set_master(11, mid("DAP"))  # DAP
    xmpu.set_master(10, mid("HSDP_DPC"))  # HSDP_DPC
    # Skip -- the middle entries for now
    xmpu.set_master(0, mid("ANY"))  # Any master


def init_xmpu(name, reg):
    base = (reg[0] << 32) | reg[1]
    size = (reg[2] << 32) | reg[3]
    nodes[name] = Xmpu(name, base, size)
    return nodes[name]


def init_instances():
    for mpu, inst in nodes.items():
        init_masters(inst)

        if debug:
            print(inst)
            print(inst.get_ctrl_reg_addr_val())
            print(inst.get_ien_reg_addr_val())
            for i, r in inst.regions.items():
                print(r)
                print(r.master)
                print(r.get_start_lo())
                print(r.get_end_lo())
                print(r.get_start_hi())
                print(r.get_end_hi())
                print(r.get_master())
                print(r.get_config())



# /*
# * Copyright (c) 2022 - 2025 Advanced Micro Devices, Inc. All Rights Reserved.
# *
# * Author:
# *       Madhav Bhatt <madhav.bhatt@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import cfgobj_hard_coding as chc

class SdtInfo:
    __sdt__ = None
    masters = {}
    ocm_high_value = None
    ocm_base_value = None
    gpos = {}
    rpu0_as_power_management_master = False
    rpu1_as_power_management_master = False
    apu_as_power_management_master = False
    rpu0_as_reset_management_master = False
    rpu1_as_reset_management_master = False
    apu_as_reset_management_master = False
    rpu0_as_overlay_config_master = False
    rpu1_as_overlay_config_master = False
    apu_as_overlay_config_master = False
    subsys_str = None
    existing_proc_type = None
    proc_type = None

    def __get_cpu_name(self, node):
        try:
            cpu_name = node["xlnx,cpu-name"].value
        except:
            cpu_name = ""
        return cpu_name

    def __is_master_defined(self, master):
        found_master = False
        subsys_str = self.subsys_str
        subsys_list = subsys_str.split("|")
        for subsys in subsys_list:
            if master in subsys.split(":")[1].split(";"):
                found_master = True
        return found_master

    def __get_masters(self):
        masters_tmp = {}
        for master in chc.masters.keys():
            for node in self.__node_tree__:
                if chc.masters[master]["name"] in  self.__get_cpu_name(node):
                    if self.__is_master_defined(master):
                        master_tmp[master] = chc.masters[master]
        if len(masters_tmp) != 0:
            self.masters = maaster_tmp
        else:
            self.masters = chc.masters
        return


    def __get_ip_name(self, node):
        try:
            ip_name = node['xlnx,ip-name'].value
        except:
            ip_name = None
        return ip_name

    def __parse_ipi_bit_pos(self):
        ip_name = []
        for master in self.masters:
            for node in self.__node_tree__:
                ip_name = self.__get_ip_name(node)
                if ip_name != None and "psu_ipi" in ip_name:
                    if self.masters[master]["name"] in self.__get_cpu_name(node):
                       self.masters[master]["ipi_bit_pos"] = node["xlnx,bit-position"].value[0]
                       self.masters[master]["is_ipi_present"] = True
                       break
        return


    def __parse_gpo_info(self):
        ip_name = []
        for node in self.__node_tree__:
            ip_name = self.__get_ip_name(node)
            if ip_name != None and "psu_pmu_iomodule" in ip_name:
                for num in chc.gpo_nums:
                    gpo_tmp = {}
                    enable_prop_str = "xlnx,gpo" + str(num) + "-enable"
                    polarity_prop_str = "xlnx,gpo" + str(num) + "-polarity"
                    try:
                        gpo_tmp["enable"] = node[enable_prop_str].value[0]
                    except:
                        gpo_tmp["enable"] = 0
                    if 1 == gpo_tmp["enable"]:
                        gpo_tmp["polarity"] = node[polarity_prop_str].value[0]
                    else:
                        gpo_tmp["polarity"] = 0
                    self.gpos["gpo" + str(num)] = gpo_tmp
                break
        return

    def __parse_ocm_base_high(self):
        ip_name = []
        for node in self.__node_tree__:
            if len(node.child_nodes) > 0:
                for child_node in node.child_nodes.values():
                    ip_name = self.__get_ip_name(child_node)
                    if ip_name != None and "psu_ocm_ram_0" in ip_name:
                        base = child_node["reg"].value[1]
                        size = child_node["reg"].value[3]
                        self.ocm_base_value = base
                        self.ocm_high_value = base + size - 1
                        break
        return

    def __parse_slaves_for_master(self):
        nodes_temp = self.__node_tree__
        nodelist = {}
        for node in nodes_temp:
            temp = node.name.split('@')
            try:
                nodelist[int(temp[1], 16)] = temp[0]
            except:
                None
        nodes = self.__sdt__.tree.nodes('/cpu.*')
        for node in nodes:
            slaves = []
            try:
                if node.propval('reg') != '':
                    if 'cpus,cluster' in node["compatible"].value:
                        cnt = 0
                        addr_cell_size = node["#ranges-address-cells"].value[0]
                        size_cell_size = node["#ranges-size-cells"].value[0]
                        while cnt < len(node["address-map"].value):
                            addr = 0
                            for cell_num in range(addr_cell_size):
                                addr = addr << 16 | node["address-map"].value[cnt + cell_num]
                            cnt += (2 * addr_cell_size) + size_cell_size + 1
                            for key in chc.node_map.keys():
                                try:
                                    for base_addr in chc.node_map[key]["base_addr"]:
                                        if base_addr == addr:
                                            periph = ""
                                            if "psu_ocm_" in chc.node_map[key]["periph"]:
                                                periph = "psu_ocm_ram_0"
                                            elif "psu_ddr" in chc.node_map[key]["periph"]:
                                                if "cpus-a53" in node.name:
                                                    periph = "psu_ddr_0"
                                                elif "cpus-r5" in  node.name:
                                                    periph = "psu_r5_ddr_0"
                                            else:
                                                periph = chc.node_map[key]["periph"]
                                            if periph not in slaves:
                                                slaves.append(periph)
                                except:
                                    None
                if node.name == "cpus-a53@0":
                    self.masters["psu_cortexa53_0"]["slaves"] = slaves
                elif node.name == "cpus-r5@0":
                    self.masters["psu_cortexr5_0"]["slaves"] = slaves
                elif node.name == "cpus-r5@1":
                    self.masters["psu_cortexr5_1"]["slaves"] = slaves
            except:
                None

    def __get_proc_type(self, options):
        try:
            self.proc_type = options['args'][1]
        except:
            self.proc_type = chc.hardcoded_proc_type
        return

    def __set_default_properties(self):
        self.rpu0_as_power_management_master = chc.rpu0_as_power_management_master
        self.rpu1_as_power_management_master = chc.rpu1_as_power_management_master
        self.apu_as_power_management_master = chc.apu_as_power_management_master
        self.rpu0_as_reset_management_master = chc.rpu0_as_reset_management_master
        self.rpu1_as_reset_management_master = chc.rpu1_as_reset_management_master
        self.apu_as_reset_management_master = chc.apu_as_reset_management_master
        self.rpu0_as_overlay_config_master = chc.rpu0_as_overlay_config_master
        self.rpu1_as_overlay_config_master = chc.rpu1_as_overlay_config_master
        self.apu_as_overlay_config_master = chc.apu_as_overlay_config_master

    def __set_properties(self, props):
        try:
            for prop in props:
                if ':' in prop:
                    key, value = prop.split(':')
                    if 'true' == value:
                        bool_value = True
                    else:
                        bool_value = False
                    if "XILPM_rpu0_as_power_management_master" == key:
                        self.rpu0_as_power_management_master = bool_value
                    elif "XILPM_rpu1_as_power_management_master" == key:
                        self.rpu1_as_power_management_master = bool_value
                    elif "XILPM_apu_as_power_management_master" == key:
                        self.apu_as_power_management_master = bool_value
                    elif "XILPM_rpu0_as_reset_management_master" == key:
                        self.rpu0_as_reset_management_master = bool_value
                    elif "XILPM_rpu1_as_reset_management_master" == key:
                        self.rpu1_as_reset_management_master = bool_value
                    elif "XILPM_apu_as_reset_management_master" == key:
                        self.apu_as_reset_management_master = bool_value
                    elif "XILPM_rpu0_as_overlay_config_master" == key:
                        self.rpu0_as_overlay_config_master = bool_value
                    elif "XILPM_rpu1_as_overlay_config_master" == key:
                        self.rpu1_as_overlay_config_master = bool_value
                    elif "XILPM_apu_as_overlay_config_master" == key:
                        self.apu_as_overlay_config_master = bool_value
        except:
            self.__set_default_proprties()
        return

    def __init__(self, sdt, options):
        self.__sdt__ = sdt
        self.__node_tree__ = list(self.__sdt__.tree)
        self.__get_proc_type(options)

        # Hard coded values
        self.__set_default_properties()
        self.__set_properties(options['args'][2:])
        self.subsys_str = chc.subsys_str

        # Parsing values from SDT
        self.__get_masters()
        self.__parse_slaves_for_master()
        self.__parse_ipi_bit_pos()
        self.__parse_ocm_base_high()
        self.__parse_gpo_info()



# /*
# * Copyright (c) 2022 - 2024 Advanced Micro Devices, Inc. All Rights Reserved.
# *
# * Author:
# *       Madhav Bhatt <madhav.bhatt@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import sys
import os
import re

sys.path.append(os.path.dirname(__file__))
sys.path.append(os.path.dirname(__file__) + "/cfg_obj/modules/")

from section import Section
import cfgobj_hard_coding as chc
from sdtinfo import SdtInfo
import cfg_data_tpl

def props():
    return ["id", "file_ext"]

def id():
    return "xlnx,output,cfgobj"

def file_ext():
    return ".c"

def is_compat(node, compat_id):
    if re.search("module,generate_config_object", compat_id):
        return cfg_obj_write
    return ""

def get_slaves_for_master(sdtinfo_obj, master):
    try:
        slave_list = sdtinfo_obj.masters[master]["slaves"]
    except:
        slave_list = []
    return slave_list

def is_rpu_lockstep(sdtinfo_obj):
    subsys_str = sdtinfo_obj.subsys_str
    found_rpu0 = False
    found_rpu1 = False

    subsys_list  = subsys_str.split("|")
    for subsys in subsys_list:
        if "RPU0" in subsys.split(":")[1].split(";"):
            found_rpu0 = True
        if "RPU1" in subsys.split(":")[1].split(";"):
            found_rpu1 = True
    if found_rpu0 == True and found_rpu1 == False:
        return True
    else:
        return False

def mask_to_str(mask):
    return "0x%08X"%mask

def is_ipi_present(master, sdtinfo_obj):
    try:
        if sdtinfo_obj.masters[master]["is_ipi_present"] == True:
            return True
        else:
            return False
    except:
        return False

def get_ipi_mask(master, sdtinfo_obj):
    if is_ipi_present(master, sdtinfo_obj) == True:
        bit_pos = sdtinfo_obj.masters[master]["ipi_bit_pos"]
        return (1<<bit_pos)
    else:
        return 0

def get_ipi_mask_txt(master, sdtinfo_obj):
    if is_ipi_present(master, sdtinfo_obj) == True:
        return "PM_CONFIG_IPI_" + master.upper() + "_MASK"
    else:
        return ""

def get_all_masters_mask_txt(sdtinfo_obj):
    macro_list = []
    for master in sdtinfo_obj.masters.keys():
        if is_ipi_present(master, sdtinfo_obj) == True:
            macro_list.append(get_ipi_mask_txt(master, sdtinfo_obj))
    if len(macro_list) > 0:
        return " | ".join(macro_list)
    else:
        return "0U"

def get_all_other_masters_mask_txt(master_name, sdtinfo_obj):
    macro_list = []
    for master in sdtinfo_obj.masters.keys():
        if (master != master_name) and (is_ipi_present(master, sdtinfo_obj) == True):
            if "psu_cortexa53_0" == master_name:
                if sdtinfo_obj.rpu0_as_power_management_master == False and \
                    "psu_cortexr5_0" == master:
                    continue
                if sdtinfo_obj.rpu1_as_power_management_master == False and \
                    False == is_rpu_lockstep(sdtinfo_obj) and "psu_cortexr5_1" == master:
                    continue
            elif "psu_cortexr5_0" == master_name:
                if sdtinfo_obj.apu_as_power_management_master == False and \
                    "psu_cortexa53_0" == master:
                    continue
                if sdtinfo_obj.rpu1_as_power_management_master == False and \
                    False == is_rpu_lockstep(sdtinfo_obj) and "psu_cortexr5_1" == master:
                    continue
            elif False == is_rpu_lockstep(sdtinfo_obj) and "psu_cortexr5_1" == master_name:
                if sdtinfo_obj.apu_as_power_management_master == False and \
                    "psu_cortexa53_0" == master:
                    continue
                if sdtinfo_obj.rpu0_as_power_management_master == False and \
                    "psu_cortexr5_0" == master:
                    continue
            macro_list.append(get_ipi_mask_txt(master, sdtinfo_obj))
    if len(macro_list) > 0:
        return " | ".join(macro_list)
    else:
        return "0U"

def get_slave_perm_mask_txt(periph, sdtinfo_obj):
    macro_list = []
    for master in sdtinfo_obj.masters.keys():
        slave_list = get_slaves_for_master(sdtinfo_obj, master)
        for slave in slave_list:
            if periph == slave and (is_ipi_present(master, sdtinfo_obj) == True):
                macro_list.append(get_ipi_mask_txt(master, sdtinfo_obj))
    if len(macro_list) > 0:
        return " | ".join(macro_list)
    else:
        return "0U"


def get_slave_perm_mask(periph, sdtinfo_obj):
    perm_mask = 0x00000000
    for master in sdtinfo_obj.masters.keys():
        slave_list = get_slaves_for_master(sdtinfo_obj, master)
        for slave in slave_list:
            if re.search(periph,slave) != None and (is_ipi_present(master, sdtinfo_obj) == True):
                perm_mask  = perm_mask | get_ipi_mask(master, sdtinfo_obj)
    return perm_mask

def get_tcm_r5_perm_mask(r5_proc, tcm_bank, sdtinfo_obj):
    perm_mask = 0x00000000
    if "psu_cortexr5_0" == r5_proc:
        if is_rpu_lockstep(sdtinfo_obj) == True:
            perm_mask = get_ipi_mask(r5_proc, sdtinfo_obj)
        else:
            if tcm_bank == "psu_r5_0_atcm_global" or tcm_bank == "psu_r5_0_btcm_global":
                perm_mask = get_ipi_mask(r5_proc, sdtinfo_obj)
    elif r5_proc == "psu_cortexr5_1":
        if is_rpu_lockstep(sdtinfo_obj) == True:
            perm_mask = 0x00000000
        else:
            if tcm_bank == "psu_r5_1_atcm_global" or tcm_bank == "psu_r5_1_btcm_global":
                perm_mask = get_ipi_mask(r5_proc, sdtinfo_obj)
    else:
        perm_mask = 0x00000000
    return perm_mask

def convert_ipi_mask_to_txt(ipi_mask, sdtinfo_obj):
    macro_list = []
    for master in sdtinfo_obj.masters.keys():
        if ((ipi_mask & get_ipi_mask(master, sdtinfo_obj)) != 0) and is_ipi_present(master, sdtinfo_obj) == True:
            macro_list.append(get_ipi_mask_txt(master, sdtinfo_obj))
    if len(macro_list) > 0:
        return " | ".join(macro_list)
    else:
        return "0U"

def get_tcm_perm_mask(tcm, sdtinfo_obj):
    perm_mask = 0x00000000
    for master in sdtinfo_obj.masters.keys():
        if re.search("psu_cortexr5_*", master) != None:
            perm_mask  = perm_mask | get_tcm_r5_perm_mask(master, tcm, sdtinfo_obj)
        else:
            slave_list = get_slaves_for_master(sdtinfo_obj, master)
            for slave in slave_list:
                if tcm == slave:
                    perm_mask  = perm_mask | get_ipi_mask(master, sdtinfo_obj)
    return perm_mask

def get_ocm_perm_mask(ocm, sdtinfo_obj):
    perm_mask = 0x00000000
    island_base = sdtinfo_obj.ocm_base_value
    island_high = sdtinfo_obj.ocm_high_value
    for master in sdtinfo_obj.masters.keys():
        if "psu_ocm_ram_0" in get_slaves_for_master(sdtinfo_obj, master):
            base_val = sdtinfo_obj.ocm_base_value
            high_val = sdtinfo_obj.ocm_high_value
            if ((island_base >= base_val) and (island_base <= high_val)) or ((island_high >= base_val) and (island_high <= high_val)):
                perm_mask = perm_mask | get_ipi_mask(master, sdtinfo_obj)
    return perm_mask

def get_mem_perm_mask(mem, sdtinfo_obj):
    perm_mask = 0x00000000
    if "psu_ddr" in mem:
        perm_mask = get_slave_perm_mask("psu_ddr_", sdtinfo_obj) | get_slave_perm_mask("psu_r5_ddr_", sdtinfo_obj)
    elif "psu_ocm_" in mem:
        perm_mask  = get_ocm_perm_mask(mem, sdtinfo_obj)
    elif re.search("psu_r5_.*tcm_global", mem) != None:
        perm_mask = get_tcm_perm_mask(mem, sdtinfo_obj)
    else:
        perm_mask = 0x00
    return perm_mask

def get_power_domain_perm_mask_txt(pwr_domain, sdtinfo_obj):
    macro_list = []
    pwr_perm_masters = chc.power_perms[pwr_domain]
    for master in pwr_perm_masters:
        if master in sdtinfo_obj.masters.keys() and is_ipi_present(master, sdtinfo_obj) == True:
            if (sdtinfo_obj.apu_as_power_management_master == False) and ("psu_cortexa53_0" == master):
                continue
            elif (sdtinfo_obj.rpu0_as_power_management_master == False) and ("psu_cortexr5_0" == master):
                continue
            elif (sdtinfo_obj.rpu1_as_power_management_master == False) and ("psu_cortexr5_1" == master):
                continue
            macro_list.append(get_ipi_mask_txt(master, sdtinfo_obj))
    if (pwr_domain == "NODE_FPD" or pwr_domain == "NODE_APU") and (len(macro_list) == 0):
        hardcode_mask = get_ipi_mask_txt("psu_cortexr5_0", sdtinfo_obj)
        if hardcode_mask != "":
            macro_list.append(hardcode_mask)
    if len(macro_list) > 0:
        return " | ".join(macro_list)
    else:
        return "0U"

def is_all_master_enabled(master_type, sdtinfo_obj):
    if "power" == master_type:
        rpu0_as_master = sdtinfo_obj.rpu0_as_power_management_master
        rpu1_as_master = sdtinfo_obj.rpu1_as_power_management_master
        apu_as_master =  sdtinfo_obj.apu_as_power_management_master
    elif "reset" == master_type:
        rpu0_as_master = sdtinfo_obj.rpu0_as_reset_management_master
        rpu1_as_master = sdtinfo_obj.rpu1_as_reset_management_master
        apu_as_master = sdtinfo_obj.apu_as_reset_management_master
    else:
        return -1
    if rpu0_as_master == True and apu_as_master == True:
        if is_rpu_lockstep(sdtinfo_obj) == False:
            if rpu1_as_master == True:
                return 1
            else:
                return 0
        else:
            return 1
    else:
        return 0


def get_list_of_management_master(master_type, sdtinfo_obj):
    macro_list = []
    if "power" == master_type:
        rpu0_as_master = sdtinfo_obj.rpu0_as_power_management_master
        rpu1_as_master = sdtinfo_obj.rpu1_as_power_management_master
        apu_as_master =  sdtinfo_obj.apu_as_power_management_master
    elif "reset" == master_type:
        rpu0_as_master = sdtinfo_obj.rpu0_as_reset_management_master
        rpu1_as_master = sdtinfo_obj.rpu1_as_reset_management_master
        apu_as_master = sdtinfo_obj.apu_as_reset_management_master
    elif "overlay_config" == master_type:
        rpu0_as_master = sdtinfo_obj.rpu0_as_overlay_config_master
        rpu1_as_master = sdtinfo_obj.rpu1_as_overlay_config_master
        apu_as_master = sdtinfo_obj.apu_as_overlay_config_master
    else:
        return "0U"
    for master in sdtinfo_obj.masters.keys():
        if is_ipi_present(master, sdtinfo_obj) != 0:
            if (apu_as_master == False) and ("psu_cortexa53_0" == master):
                continue
            elif (rpu0_as_master == False) and ("psu_cortexr5_0" == master):
                continue
            elif (rpu1_as_master == False) and ("psu_cortexr5_1" == master):
                continue
            macro_list.append(get_ipi_mask_txt(master, sdtinfo_obj))
    if len(macro_list) > 0:
        return " | ".join(macro_list)
    else:
        return "0U"

def generate_master_ipi_mask_def(sdtinfo_obj):
    out_lines = ["\n"]
    for master in sdtinfo_obj.masters.keys():
        if is_ipi_present(master, sdtinfo_obj) == True:
            out_lines.append("#define " + get_ipi_mask_txt(master, sdtinfo_obj) + "    " + mask_to_str(get_ipi_mask(master, sdtinfo_obj)) + "\n")
    out_lines.append("\n\n")
    return out_lines

def get_prealloc_for_master_txt(master_name, prealloc_list, sdtinfo_obj):
    node_count = 0
    master_prealloc_txt = []
    if is_ipi_present(master_name, sdtinfo_obj) == True:
        master_mask = get_ipi_mask_txt(master_name, sdtinfo_obj)
        master_prealloc_txt.append("\t/* Prealloc for " + master_name + " */\n")
        master_prealloc_txt.append("\t" + master_mask + ",\n")
        for key in prealloc_list:
            periph_perms = chc.node_map[key]['perms']
            periph_name = chc.node_map[key]['periph']
            periph_type = chc.node_map[key]['type']
            periph_label = chc.node_map[key]['label']
            if master_mask in periph_perms:
                master_prealloc_txt.append("\t" + periph_label + ",\n")
                master_prealloc_txt.append("\tPM_MASTER_USING_SLAVE_MASK, /* Master is using Slave */\n")
                master_prealloc_txt.append("\tPM_CAP_ACCESS | PM_CAP_CONTEXT, /* Current Requirements */\n")
                master_prealloc_txt.append("\tPM_CAP_ACCESS | PM_CAP_CONTEXT, /* Default Requirements */\n")
                master_prealloc_txt.append("\n")
                node_count += 1
    master_prealloc_txt.insert(2, "\t" + str(node_count) + ",\n")
    master_prealloc_txt.append("\n")
    return master_prealloc_txt


def generate_master_section_data(sdtinfo_obj):
    out_lines = []
    out_lines.append("\tPM_CONFIG_MASTER_SECTION_ID, /* Master SectionID */" + "\n")
    master_count = len(sdtinfo_obj.masters.keys())
    out_lines.append("\t" + str(master_count) + "U, /* No. of Masters*/" + "\n")
    out_lines.append("\n")
    for master in sdtinfo_obj.masters.keys():
        if sdtinfo_obj.masters[master]["name"] == "RPU0":
            if is_rpu_lockstep(sdtinfo_obj) == True:
                master_node = "NODE_RPU"
            else:
                master_node = "NODE_RPU_0"
        elif sdtinfo_obj.masters[master]["name"] == "RPU1":
            master_node = "NODE_RPU_1"
        elif sdtinfo_obj.masters[master]["name"] == "APU":
            master_node = "NODE_APU"
        out_lines.append("\t" + master_node + ", /* Master Node ID */" + "\n")
        if is_ipi_present(master, sdtinfo_obj) == True:
            out_lines.append("\t" + get_ipi_mask_txt(master, sdtinfo_obj) + ", /* IPI Mask of this master */" + "\n")
        else:
            out_lines.append("\t0U, /* IPI Mask of this master */" + "\n")
        out_lines.append("\tSUSPEND_TIMEOUT, /* Suspend timeout */" + "\n")
        out_lines.append("\t" +  get_all_other_masters_mask_txt(master, sdtinfo_obj) + ", /* Suspend permissions */" + "\n")
        out_lines.append("\t" +  get_all_other_masters_mask_txt(master, sdtinfo_obj) + ", /* Wake permissions */" + "\n")
        out_lines.append("\n")
    out_lines.append("\n")
    return out_lines

def generate_slave_section_data(sdtinfo_obj):
    out_lines = ["\n\n\tPM_CONFIG_SLAVE_SECTION_ID,\t/* Section ID */\n"]
    slave_count = 0
    for key in chc.node_map:
        periph_name = chc.node_map[key]['periph']
        periph_type = chc.node_map[key]['type']
        periph_label = chc.node_map[key]['label']
        if periph_type == "slave":
            chc.node_map[key]['perms'] = get_slave_perm_mask_txt(periph_name, sdtinfo_obj)
        elif (periph_type == "memory") and (periph_type != "NA"):
            chc.node_map[key]['perms'] = convert_ipi_mask_to_txt(get_mem_perm_mask(periph_name, sdtinfo_obj), sdtinfo_obj)
        elif periph_type == "others":
            chc.node_map[key]['perms'] = get_all_masters_mask_txt(sdtinfo_obj)

        if ("slave" == periph_type) or ("memory" == periph_type and "NA" != periph_type) or ("others" == periph_type):
            if chc.node_map[key]['perms'] == "0U":
                continue
            slave_count += 1
            out_lines.append("\t" + periph_label + ",\n")
            out_lines.append("\tPM_SLAVE_FLAG_IS_SHAREABLE,\n")
            out_lines.append("\t" + chc.node_map[key]['perms']+ ", /* IPI Mask */\n\n")

        ipi_perm = ""
        if periph_type == "ipi":
            if periph_label == "NODE_IPI_APU":
                if "psu_cortexa53_0" in sdtinfo_obj.masters.keys() and is_ipi_present("psu_cortexa53_0", sdtinfo_obj) == True:
                    ipi_perm = get_ipi_mask_txt("psu_cortexa53_0", sdtinfo_obj)
                else:
                    ipi_perm = ""
            elif periph_label == "NODE_IPI_RPU_0":
                if "psu_cortexr5_0" in sdtinfo_obj.masters.keys() and is_ipi_present("psu_cortexr5_0", sdtinfo_obj) == True:
                    ipi_perm = get_ipi_mask_txt("psu_cortexr5_0", sdtinfo_obj)
                else:
                    ipi_perm = ""
            elif periph_label == "NODE_IPI_RPU_1":
                if "psu_cortexr5_1" in sdtinfo_obj.masters.keys() and is_ipi_present("psu_cortexr5_1", sdtinfo_obj) == True:
                    ipi_perm = get_ipi_mask_txt("psu_cortexr5_1", sdtinfo_obj)
                else:
                    ipi_perm = ""
            else:
                ipi_perm = ""
            if ipi_perm != "":
                slave_count += 1
                chc.node_map[key]['perms'] = ipi_perm
                out_lines.append("\t" + periph_label + ",\n")
                out_lines.append("\t0U,\n")
                out_lines.append("\t" + chc.node_map[key]['perms']+ ", /* IPI Mask */\n\n")

    out_lines.insert(1, "\t" + str(slave_count) + ",\t\t\t\t/* Number of slaves */\n\n")
    out_lines.append("\n")
    return out_lines

def generate_prealloc_section_data(sdtinfo_obj):
    out_lines = ["\n"]
    master_count = 0
    proc_type = sdtinfo_obj.proc_type
    if is_ipi_present("psu_cortexa53_0", sdtinfo_obj) == True:
        chc.apu_prealloc_list.append("NODE_IPI_APU")
    if proc_type == "psu_cortexr5_0":
        chc.rpu_0_prealloc_list.extend(chc.rpu_0_prealloc_conditional_list)
    if is_ipi_present("psu_cortexr5_0", sdtinfo_obj) == True:
        chc.rpu_0_prealloc_list.append("NODE_IPI_RPU_0")
    if is_ipi_present("psu_cortexr5_1", sdtinfo_obj) == True:
        chc.rpu_1_prealloc_list.append("NODE_IPI_RPU_1")
    out_lines.append("\tPM_CONFIG_PREALLOC_SECTION_ID, /* Preallaoc SectionID */\n")
    for master in sdtinfo_obj.masters.keys():
        if is_ipi_present(master, sdtinfo_obj) == True:
            master_count += 1
    out_lines.append("\t" + str(master_count) + "U, /* No. of Masters*/\n")
    out_lines.append("\n")
    if "psu_cortexa53_0" in sdtinfo_obj.masters.keys() and is_ipi_present("psu_cortexa53_0", sdtinfo_obj):
        out_lines.extend(get_prealloc_for_master_txt("psu_cortexa53_0", chc.apu_prealloc_list, sdtinfo_obj))
    if "psu_cortexr5_0" in sdtinfo_obj.masters.keys() and is_ipi_present("psu_cortexr5_0", sdtinfo_obj):
        out_lines.extend(get_prealloc_for_master_txt("psu_cortexr5_0", chc.rpu_0_prealloc_list, sdtinfo_obj))
    if "psu_cortexr5_1" in sdtinfo_obj.masters.keys() and is_ipi_present("psu_cortexr5_1", sdtinfo_obj):
        out_lines.extend(get_prealloc_for_master_txt("psu_cortexr5_1", chc.rpu_1_prealloc_list, sdtinfo_obj))
    out_lines.append("\t\n")
    return out_lines

def generate_power_section_data(sdtinfo_obj):
    out_lines = ["\n"]
    out_lines.append("\tPM_CONFIG_POWER_SECTION_ID, /* Power Section ID */\n")
    out_lines.append("\t" + str(len(chc.power_node_list)) + "U, /* Number of power nodes */\n")
    out_lines.append("\n")
    for node in chc.power_node_list:
        out_lines.append("\t" + node + ", /* Power node ID */\n")
        out_lines.append("\t" + get_power_domain_perm_mask_txt(node, sdtinfo_obj) + ", /* Force power down permissions */\n")
        out_lines.append("\n")
    out_lines.append("\n")
    return out_lines

def get_periph_perm_mask_txt_for_rst_line(reset_line, sdtinfo_obj):
    macro_list1 = []
    macro_list2 = []
    reset_management_master_list = get_list_of_management_master("reset", sdtinfo_obj)

    line_node = chc.reset_line_map[reset_line]['node']
    periph_name = chc.node_map[line_node]['periph']
    periph_type = chc.node_map[line_node]['type']

    if periph_type == 'slave':
        for master in sdtinfo_obj.masters.keys():
            slave_list = get_slaves_for_master(sdtinfo_obj, master)
            for slave in slave_list:
                if periph_name == slave and (is_ipi_present(master, sdtinfo_obj) == True):
                    master_ipi_mask_txt = get_ipi_mask_txt(master, sdtinfo_obj)
                    macro_list1.append(master_ipi_mask_txt)
                    if master_ipi_mask_txt in reset_management_master_list:
                        macro_list2.append(master_ipi_mask_txt)
    elif periph_type == 'memory':
        if periph_name == 'psu_ddr':
            mem_perms = get_mem_perm_mask(periph_name, sdtinfo_obj)
        elif periph_name == 'psu_ocm_0':
            ocm_bank_0_perms = get_mem_perm_mask('psu_ocm_0', sdtinfo_obj)
            ocm_bank_1_perms = get_mem_perm_mask('psu_ocm_1', sdtinfo_obj)
            ocm_bank_2_perms = get_mem_perm_mask('psu_ocm_2', sdtinfo_obj)
            ocm_bank_3_perms = get_mem_perm_mask('psu_ocm_3', sdtinfo_obj)
            mem_perms = ocm_bank_0_perms | ocm_bank_1_perms | ocm_bank_2_perms | ocm_bank_3_perms
        else:
            mem_perms = 0
        for master in sdtinfo_obj.masters.keys():
            if (mem_perms & get_ipi_mask(master, sdtinfo_obj)) != 0 and (is_ipi_present(master, sdtinfo_obj)) == True:
                master_ipi_mask_txt = get_ipi_mask_txt(master, sdtinfo_obj)
                macro_list1.append(master_ipi_mask_txt)
                if master_ipi_mask_txt in reset_management_master_list:
                    macro_list2.append(master_ipi_mask_txt)

    if macro_list2:
        return " | ".join(macro_list2)
    elif macro_list1:
        return " | ".join(macro_list1)
    else:
        return "0U"

def generate_reset_section_data(sdtinfo_obj):
    out_lines = ["\n"]
    reset_management_master_list = get_list_of_management_master("reset", sdtinfo_obj)
    out_lines.append("\tPM_CONFIG_RESET_SECTION_ID, /* Reset Section ID */\n")
    out_lines.append("\t" + str(len(chc.reset_line_map)) + "U, /* Number of resets */\n")
    out_lines.append("\n")
    for reset_line in chc.reset_line_map:
        line_name = chc.reset_line_map[reset_line]["label"]
        line_type = chc.reset_line_map[reset_line]["type"]
        if line_type == "normal":
            out_lines.append("\t" + line_name + ", " + get_all_masters_mask_txt(sdtinfo_obj) + ",\n")
        elif line_type == "rpu_only":
            if ("psu_cortexr5_0" in sdtinfo_obj.masters.keys()) and (is_ipi_present("psu_cortexr5_0", sdtinfo_obj) == True):
                out_lines.append("\t" + line_name + ", " + get_ipi_mask_txt("psu_cortexr5_0", sdtinfo_obj) + ",\n")
            else:
                out_lines.append("\t" + line_name + ", 0,\n")
        elif (1 == is_all_master_enabled("reset", sdtinfo_obj)) and ((line_type == "rst_periph") or (line_type == "rst_shared" ) or (line_type == "rst_proc")) :
            out_lines.append("\t" + line_name + ", " + get_all_masters_mask_txt(sdtinfo_obj) + ",\n")
        elif (0 == is_all_master_enabled("reset", sdtinfo_obj)) and ((line_type == "rst_periph") or (line_type == "rst_shared" ) or (line_type == "rst_proc")) :
            if line_type == "rst_periph":
                perms = get_periph_perm_mask_txt_for_rst_line(reset_line, sdtinfo_obj)
                out_lines.append("\t" + line_name + ", " + perms + ",\n")
            elif line_type == "rst_shared":
                out_lines.append("\t" + line_name + ", " + reset_management_master_list + ",\n")
            elif line_type == "rst_proc":
                line_proc = chc.reset_line_map[reset_line]["proc"]
                macro_list  = ""
                master_txt = ""
                if line_proc == "APU":
                    master_txt += get_ipi_mask_txt("psu_cortexa53_0", sdtinfo_obj)
                elif (((line_proc == "RPU_1") or (line_proc == "RPU")) and is_rpu_lockstep(sdtinfo_obj)) or line_proc == "RPU_0" :
                    master_txt += get_ipi_mask_txt("psu_cortexr5_0", sdtinfo_obj)
                elif line_proc == "RPU_1":
                    master_txt += get_ipi_mask_txt("psu_cortexr5_1", sdtinfo_obj)
                elif line_proc == "RPU":
                    master_rpu_0 = get_ipi_mask_txt("psu_cortexr5_0", sdtinfo_obj)
                    master_rpu_1 = get_ipi_mask_txt("psu_cortexr5_1", sdtinfo_obj)
                    if ((master_rpu_0 not in reset_management_master_list) and len(master_rpu_0) > 0) and \
                       ((master_rpu_1 not in reset_management_master_list) and len(master_rpu_1) > 0):
                        master_txt = master_txt + master_rpu_0 + " | " + master_rpu_1
                if (master_txt not in reset_management_master_list) and len(master_txt) > 0:
                    macro_list += master_txt
                if "0U" != reset_management_master_list:
                    if len(macro_list) > 0:
                        macro_list += " | "
                    macro_list += reset_management_master_list
                if len(macro_list) == 0:
                    macro_list += "0U"
                out_lines.append("\t" + line_name + ", " + macro_list + ",\n")
        else:
            out_lines.append("\t" + line_name + ", 0,\n")
    out_lines.append("\n")
    return out_lines

def generate_set_config_section_data(sdtinfo_obj):
    out_lines = []
    overlay_config_master_list = get_list_of_management_master("overlay_config", sdtinfo_obj)
    out_lines.append("\tPM_CONFIG_SET_CONFIG_SECTION_ID,\t\t/* Set Config Section ID */\n")
    out_lines.append("\t0U, /* Permissions to load base config object */\n")
    out_lines.append("\t" + overlay_config_master_list + ", /* Permissions to load overlay config object */\n")
    out_lines.append("\n")

    return out_lines

def generate_shutdown_section_data(sdtinfo_obj):
    out_lines = ["\n"]
    power_management_master_list = get_list_of_management_master("power", sdtinfo_obj)
    out_lines.append("\tPM_CONFIG_SHUTDOWN_SECTION_ID, /* Shutdown Section ID */\n")
    out_lines.append("\t" + power_management_master_list + ", /* System Shutdown/Restart Permission */\n")
    out_lines.append("\n")
    return out_lines

def generate_gpo_section_data(sdtinfo_obj):
    out_lines = []
    out_lines.append("\tPM_CONFIG_GPO_SECTION_ID,\t\t/* GPO Section ID */\n")
    for num in chc.gpo_nums:
        # Validate rows before querying
        if "gpo" + str(num) not in sdtinfo_obj.gpos:
            continue
        if 1 == sdtinfo_obj.gpos["gpo" + str(num)]["polarity"]:
            out_lines.append("\tPM_CONFIG_GPO1_BIT_" + str(num) + "_MASK |\n")
        if 1 == sdtinfo_obj.gpos["gpo" + str(num)]["enable"]:
            out_lines.append("\tPM_CONFIG_GPO1_MIO_PIN_"+ str(32+num) +"_MAP |\n")
    out_lines.append("\t0,\t\t\t\t\t/* State of GPO pins */")
    out_lines.append("\n")
    return out_lines

def generate_tpl_lines():
    final_lines = cfg_data_tpl.config_object_template.split('\n')
    for line_num in range(len(final_lines)):
        final_lines[line_num] += '\n'
    return final_lines

def cfg_obj_write(root_node, sdt, options):
    if options.get('outdir', {}):
        sdt.outdir = options['outdir']
    sdtinfo_obj = SdtInfo(sdt, options)

    final_lines = generate_tpl_lines()

    sections = [
                {"identifier" : "<<MASTER_IPI_MASK_DEF>>",     "handler" : generate_master_ipi_mask_def},
                {"identifier" : "<<MASTER_SECTION_DATA>>",     "handler" : generate_master_section_data},
                {"identifier" : "<<SLAVE_SECTION_DATA>>",      "handler" : generate_slave_section_data},
                {"identifier" : "<<PREALLOC_SECTION_DATA>>",   "handler" : generate_prealloc_section_data},
                {"identifier" : "<<POWER_SECTION_DATA>>",      "handler" : generate_power_section_data},
                {"identifier" : "<<RESET_SECTION_DATA>>",      "handler" : generate_reset_section_data},
                {"identifier" : "<<SET_CONFIG_SECTION_DATA>>", "handler" : generate_set_config_section_data},
                {"identifier" : "<<SHUTDOWN_SECTION_DATA>>",   "handler" : generate_shutdown_section_data},
                {"identifier" : "<<GPO_SECTION_DATA>>",        "handler" : generate_gpo_section_data},
               ]

    for section in sections:
        section_obj =  Section(section["identifier"], section["handler"], sdtinfo_obj)
        final_lines = section_obj.replace_section(final_lines)

    outfile_name = options["args"][0]
    outfile = os.path.join(sdt.outdir, outfile_name)
    outfile = open(outfile, 'w')
    outfile.writelines(final_lines)
    return True



#/*
# * Copyright (c) 2022 AMD Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import sys
import types
import os
import getopt
import re
import copy
import subprocess
import textwrap

def is_compat( node, compat_string_to_test ):
    if re.search( "module,image-builder", compat_string_to_test):
        return image_builder
    return ""

def usage():
    print( """
   Usage: image-builder [--uboot] -o <output dir> --imagebuilder <path to imagebuilder>

    wrapper around imagebuilder (https://gitlab.com/xen-project/imagebuilder)      

      --uboot  execute imagebuilder's "uboot-script-gen", with the
               options: -t tftp -c ./config, and the supplied output directory
      -i       path to imagebuilder clone
      -v       enable verbose debug/processing
      -o       output directory for files

    """)


def image_builder( tgt_node, sdt, options ):
    try:
        verbose = options['verbose']
    except:
        verbose = 0

    try:
        args = options['args']
    except:
        args = []

    try:
        xen_tree = sdt.subtrees["extracted"]
    except:
        print( "[ERROR]: no extracted tree detected, returning" )
        return False

    opts,args2 = getopt.getopt( args, "vpt:o:i:", [ "uboot", "verbose", "imagebuilder=" ] )

    image_type="uboot"
    output=None
    for o,a in opts:
        # print( "o: %s a: %s" % (o,a))
        if o in ('-o'):
            output=a
        elif o in ('-v', "--verbose"):
            verbose = verbose + 1
        elif o in ("--uboot"):
            image_type="uboot"
        elif o in ("-i", "--imagebuilder"):
            image_builder=a

    if not image_builder:
        print( "[ERROR][imagebuilder]: path to image builder not passed" )
        sys.exit(1)

    if image_type == "uboot":
        print( "[INFO][imagebuilder]: generating uboot" )

        if not output:
            print( "[ERROR]: path to imagebuilder missing" )
            sys.exit(1)

        #  ~/git/imagebuilder/scripts/uboot-script-gen -t tftp -d . -c ./config
        result = subprocess.run(['%s/scripts/uboot-script-gen' % image_builder, '-t', 'tftp', '-d', '%s' % output, '-c', '%s/config' % output ], stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=3)
        if result.returncode != 0:
            print( "[ERROR]: unable to generate uboot scripts" )
            print( "\n%s" % textwrap.indent(result.stderr.decode(), '         ') )
            sys.exit(result.returncode)

    return True



#/*
# * Copyright (c) 2020 -2026 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.rao@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import sys
import os
import re
import lopper_lib
import common_utils as utils

from baremetalconfig_xlnx import scan_reg_size, get_cpu_node, get_label
from common_utils import to_cmakelist
from openamp_xlnx import xlnx_openamp_get_ddr_elf_load
from lopper.log import _init, _warning, _info, _error, _debug, _level

sys.path.append(os.path.dirname(__file__))

_init(__name__)

def is_compat( node, compat_string_to_test ):
    if re.search( "module,baremetallinker_xlnx", compat_string_to_test):
        return xlnx_generate_bm_linker
    return ""


# tgt_node: is the baremetal config top level domain node number
# sdt: is the system device-tree
# options: baremetal application source path
def get_memranges(tgt_node, sdt, options):
    _level(utils.log_setup(options), __name__)
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()
    mem_nodes = []

    #Maintain a static memory IP list this is needed inorder to capture proper ip name in the linker script
    xlnx_memipname = {
        "axi_bram": 0, "ps7_ddr": 0, "psu_ddr": 0,
        "psv_ddr": 0, "mig": 0, "lmb_bram": 0,
        "axi_noc2": 0, "axi_noc": 0,"psu_ocm": 0,
        "psv_ocm": 0, "psx_ocm": 0, "ocm": 0, "ddr4": 0,
        "ddr5": 0, "mig_7series": 0, "ps7_ram": 0,
        "axi_emc": 0, "psu_qspi_linear": 0, "ps7_qspi_linear": 0, "pmc_ram": 0,
        "ddr3": 0, "lpddrmc" : 0, "lpddr": 0, "axi_xspi": 0, "psv_xram": 0
    }
    symbol_node = ""
    for node in root_sub_nodes:
        try:
            if node.name == "__symbols__":
                symbol_node = node
            device_type = node["device_type"].value
            if "memory" in device_type:
                mem_nodes.append(node)
        except:
           pass

    zynqmp_fsbl = None
    try:
        if options['args'][2] == "zynqmp_fsbl":
            zynqmp_fsbl = 1
    except:
        pass

    xparam = False
    try:
        if "xparam" in options['args']:
            xparam = True
    except:
        pass

    # Ensure that the region addresses are always in descending order of addresses
    # This order is necessary to employ the comparison while mapping the available regions.
    versal_noc_region_ranges =  {
        "0x70000000000": "DDR_CH_3", "0x60000000000": "DDR_CH_2", "0x50000000000": "DDR_CH_1",
        "0x20000000000": "DDR_PL_NOC", "0x10000000000": "DDR_LOW_3", "0xC000000000": "DDR_LOW_2",
        "0x47C0000000": "HBM15_PC1", "0x4780000000": "HBM15_PC0", "0x4740000000": "HBM14_PC1", "0x4700000000": "HBM14_PC0",
        "0x46C0000000": "HBM13_PC1", "0x4680000000": "HBM13_PC0", "0x4640000000": "HBM12_PC1", "0x4600000000": "HBM12_PC0",
        "0x45C0000000": "HBM11_PC1", "0x4580000000": "HBM11_PC0", "0x4540000000": "HBM10_PC1", "0x4500000000": "HBM10_PC0",
        "0x44C0000000": "HBM9_PC1", "0x4480000000": "HBM9_PC0", "0x4440000000": "HBM8_PC1", "0x4400000000": "HBM8_PC0",
        "0x43C0000000": "HBM7_PC1", "0x4380000000": "HBM7_PC0", "0x4340000000": "HBM6_PC1", "0x4300000000": "HBM6_PC0",
        "0x42C0000000": "HBM5_PC1", "0x4280000000": "HBM5_PC0", "0x4240000000": "HBM4_PC1", "0x4200000000": "HBM4_PC0",
        "0x41C0000000": "HBM3_PC1", "0x4180000000": "HBM3_PC0", "0x4140000000": "HBM2_PC1", "0x4100000000": "HBM2_PC0",
        "0x40C0000000": "HBM1_PC1", "0x4080000000": "HBM1_PC0", "0x4040000000": "HBM0_PC1", "0x4000000000": "HBM0_PC0",
        "0x800000000": "DDR_LOW_1",
        "0x0": "DDR_LOW_0"
    }

    # Ensure that the region addresses are always in descending order of addresses
    # This order is necessary to employ the comparison while mapping the available regions.
    versal_net_noc2_region_ranges = {
        "0x198000000000" : "DDR_CH_4",
        "0x190000000000" : "DDR_CH_3A",
        "0x188000000000" : "DDR_CH_3",
        "0x180000000000" : "DDR_CH_2A",
        "0x70000000000" : "DDR_CH_2",
        "0x60000000000" : "DDR_CH_1A",
        "0x50000000000" : "DDR_CH_1",
        "0x10000000000" : "DDR_LOW_3",
        "0xC000000000" : "DDR_LOW_2",
        "0x800000000" : "DDR_LOW_1",
        "0x0" : "DDR_LOW_0"
    }

    noc_regions = versal_noc_region_ranges

    # Yocto Machine to CPU compat mapping
    match_cpunode = get_cpu_node(sdt, options)
    if not match_cpunode:
        return
    address_map = match_cpunode.parent["address-map"].value
    all_phandles = {}
    ns = match_cpunode.parent["#ranges-size-cells"].value[0]
    na = match_cpunode.parent["#ranges-address-cells"].value[0]
    cells = na + ns
    tmp = na
    while tmp < len(address_map):
        all_phandles[tmp] = address_map[tmp]
        tmp = tmp + cells + na + 1

    mem_ranges = {}
    lable_names = {}
    # Remove Duplicate memory node referenecs
    mem_nodes = list(dict.fromkeys(mem_nodes))
    for node in mem_nodes:
        # Check whether the memory node is mapped to cpu cluster or not
        indx_list = [index for index,handle in all_phandles.items() if handle == node.phandle]
        addr_list = []
        size_list = []
        fsbl_update_size = False
        label_name = get_label(sdt, symbol_node, node)
        if zynqmp_fsbl and "psu_ddr_0_memory" in label_name:
            val = node['reg'].value
            nac = node.parent["#address-cells"].value[0]
            nsc = node.parent["#size-cells"].value[0]
            total_nodes = int(len(val)/(nac+nsc))
            size = 0
            for i in range(total_nodes):
                reg, sz = scan_reg_size(node, val, i)
                size += sz
            fsbl_update_size = size

        for inx in indx_list:
            start = [address_map[inx+i+1] for i in range(na)]
            size_list.append(address_map[inx+2*na])
            if na == 2 and start[0] != 0:
                reg = int(f"{hex(start[0])}{start[1]:08x}", base=16)
                addr_list.append(reg)
            elif na == 2:
                addr_list.append(start[1])
            else:
                addr_list.append(start[0])

        nac = node.parent["#address-cells"].value[0]
        nsc = node.parent["#size-cells"].value[0]
        val = node['reg'].value
        total_nodes = int(len(val)/(nac+nsc))
        name_list = [name.replace("_", "-") for name in list(xlnx_memipname.keys())]
        try:
            compat = node.propval('compatible')[0]
            match = [mem for mem in name_list if mem in compat]
            for i in range(total_nodes):
                reg, size = scan_reg_size(node, val, i)
                valid_range = [addr for addr in addr_list if reg == addr or addr in range(reg, size-reg)]
                if not valid_range:
                    valid_range = [reg for index, addr in enumerate(addr_list) if reg in range(addr, size_list[index]-addr)]
                if valid_range:
                    key = match[0].replace("-", "_")
                    is_valid_noc_ch = 0
                    if "axi_noc" in key:
                        if "axi_noc2" in key:
                            family = sdt.tree['/'].propval('family')[0]
                            if family != "Versal":
                                noc_regions = versal_net_noc2_region_ranges
                        for region_addr_range in noc_regions.keys():
                            if int(region_addr_range, base=16) <= int(hex(valid_range[0]), base=16):
                                is_valid_noc_ch = noc_regions[region_addr_range]
                                break

                    if is_valid_noc_ch:
                        linker_secname = key + str("_") + is_valid_noc_ch
                        label_name_new = label_name + str("_") + is_valid_noc_ch
                    else:
                        linker_secname = key + str("_") + str(xlnx_memipname[key])
                        label_name_new = label_name + str("_") + str(xlnx_memipname[key])
                        xlnx_memipname[key] += 1

                    if not xparam and (linker_secname in mem_ranges):
                        index_val = str("_") + str(xlnx_memipname[key])
                        linker_secname += index_val
                        label_name_new += index_val
                        xlnx_memipname[key] += 1

                    lable_names[linker_secname] = label_name_new
                    # Update the mem_ranges to account for multiple NoC memory segments within a given region.
                    if is_valid_noc_ch and xparam and (linker_secname in mem_ranges):
                        start_addr, old_size = mem_ranges[linker_secname]
                        new_size = valid_range[0] + size - start_addr
                        mem_ranges.update({linker_secname: [start_addr, new_size]})
                        continue
                    if fsbl_update_size:
                        mem_ranges.update({linker_secname: [valid_range[0], fsbl_update_size]})
                    else:
                        mem_ranges.update({linker_secname: [valid_range[0], size]})
        except KeyError:
            pass
    return mem_ranges, lable_names


def get_unique_memip_list(mem_ranges):
    unique_mem_ip_list = []
    for memip in mem_ranges.keys():
        ip_name = memip.rsplit("_")[0]
        if ip_name not in unique_mem_ip_list:
            unique_mem_ip_list += [ip_name]
    return unique_mem_ip_list


# tgt_node: is the baremetal config top level domain node number
# sdt: is the system device-tree
# options: baremetal application source path
def xlnx_generate_bm_linker(tgt_node, sdt, options):
    _level(utils.log_setup(options), __name__)
    mem_ranges,lable_names = get_memranges(tgt_node, sdt, options)
    openamp_config = None
    memtest_config = None
    machine = options['args'][0]
    cpu_ip_name = None
    default_ddr = None
    memip_list = []
    match_cpunode = get_cpu_node(sdt, options)
    if match_cpunode.propval('xlnx,ip-name') != ['']:
        cpu_ip_name = match_cpunode.propval('xlnx,ip-name', list)[0]
    try:
        extra_config = options['args'][2]
        if extra_config == "memtest":
            memtest_config = True
        if extra_config == "openamp":
            openamp_config = True

    except IndexError:
        pass

    src_dir = options['args'][1]
    app_path = utils.get_dir_path(src_dir.rstrip(os.sep))
    appname = utils.get_base_name(app_path)
    yaml_file = os.path.join(app_path, "data", f"{appname}.yaml")

    cmake_file = os.path.join(sdt.outdir, f"{appname.capitalize()}Example.cmake")
    cfd = open(cmake_file, 'a')
    traverse  = False if memtest_config else True
    default_ddr,memtest_config, mb_reset_addr = get_ddr_address(sdt,tgt_node,mem_ranges,match_cpunode,cpu_ip_name,memtest_config,traverse)

    openamp_elfload_start = False
    openamp_elfload_sz = False
    openamp_entry_name = False
    if openamp_config:
        elfload_tuple = xlnx_openamp_get_ddr_elf_load(machine, sdt)
        if elfload_tuple != None:
            openamp_elfload_start = elfload_tuple[0]
            openamp_elfload_sz = elfload_tuple[1]
            openamp_entry_name = elfload_tuple[2]

    memip_list = []
    for key, value in sorted(mem_ranges.items(), key=lambda e: e[1][1], reverse=traverse):
        # Here skip adding DDR system mem to config if OpenAMP module is set
        if "psu_qspi_linear" in key or "ps7_qspi_linear" in key or "axi_emc" in key or openamp_elfload_sz:
            continue
        start,size = value[0], value[1]
        """
        LMB BRAM initial 80 bytes being used by the linker vectors section in case of Microblaze
        Adjust the size and start address accordingly.
        """
        if "lmb_bram" in key and not "microblaze_riscv" in machine and not "cortex" in machine:
            start += 80
            size -= start
        memip_list.append(key)
        cfd.write("set(%s %s)\n" % (lable_names[key], to_cmakelist([hex(start), hex(size)])))

    if memtest_config:
        default_ddr, memip_list = generate_mem_config(mem_ranges,memtest_config,memip_list,default_ddr,cpu_ip_name)
    if mb_reset_addr:
        cfd.write("set(BASE_VECTOR %s)\n" % mb_reset_addr)
    cfd.write("set(DDR %s)\n" % lable_names[default_ddr])
    cfd.write("set(CODE %s)\n" % lable_names[default_ddr])
    cfd.write("set(DATA %s)\n" % lable_names[default_ddr])
    memip_list = [lable_names[i] for i in memip_list]
    cfd.write("set(TOTAL_MEM_CONTROLLERS %s)\n" % to_cmakelist(memip_list))

    generate_linker_script(mem_ranges,yaml_file,appname,cfd,cpu_ip_name,machine,openamp_config,openamp_elfload_sz,openamp_elfload_start,openamp_entry_name,traverse,lable_names,mb_reset_addr)

    return True

def generate_linker_script(mem_ranges,yaml_file,appname,cfd,cpu_ip_name,machine,openamp_config,openamp_elfload_sz,openamp_elfload_start,openamp_entry_name,traverse,lable_names,mb_reset_addr):
    """
    To get generate the linker script
    Parameters:
        mem_ranges(list)           : List of mem ranges
        yaml_file(path)            : Yaml file path
        appname(str)               : Name of the App
        cfd(file obj)              : Cmake file object
        cpu_ip_name(str)           : Cpu ip name
        machine(str)               : Machine name
        openamp_config(bool)       : True or False
        openamp_elfload_sz(int)    : Openamp elf load size
        openamp_elfload_start(int) : Openamp elf start
        openamp_entry_name(str)    : OpenAMP elf entry name
        traverse(bool)             : True or False
        lable_names(list)          : Lable names
    Return:
        None
    """
    stack_size = None
    heap_size = None
    user_heap_size = None
    if cpu_ip_name is not None:
        if "microblaze" in cpu_ip_name:
            stack_size = 0x400
            heap_size = 0x800
        else:
            stack_size = 0x2000
            heap_size = 0x2000
    if not utils.is_file(yaml_file):
        _warning(f"{appname} doesn't have yaml file")
    else:
        schema = utils.load_yaml(yaml_file)
        if schema.get("linker_constraints"):
            linker_constraint_opts = list(schema["linker_constraints"].keys())
            if "stack" in linker_constraint_opts:
                stack_size = schema["linker_constraints"]["stack"]
            if "heap" in linker_constraint_opts:
                heap_size = schema["linker_constraints"]["heap"]
                user_heap_size = True


    mem_sec = ""
    """
    For cortexr5 processor in ZynqMP and Versal SOC TCM memory map is fixed
    update the memory section for the same.
    """
    if "psu_cortexr5" in machine:
        mem_sec += '\n\tpsu_r5_0_atcm_MEM_0 : ORIGIN = 0x0, LENGTH = 0x10000'
        mem_sec += '\n\tpsu_r5_0_btcm_MEM_0 : ORIGIN = 0x20000, LENGTH = 0x10000'
        mem_sec += '\n\tpsu_r5_tcm_ram_0_MEM_0 : ORIGIN = 0x0, LENGTH = 0x40000'
    if "psv_cortexr5" in machine:
        mem_sec += '\n\tpsv_r5_0_atcm_MEM_0 : ORIGIN = 0x0, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_0_atcm_lockstep_MEM_0 : ORIGIN = 0xFFE10000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_0_btcm_MEM_0 : ORIGIN = 0x20000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_0_btcm_lockstep_MEM_0 : ORIGIN = 0xFFE30000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_pmc_ram_psv_pmc_ram : ORIGIN = 0xF2000000, LENGTH = 0x20000'
        mem_sec += '\n\tpsv_r5_0_data_cache_MEM_0 : ORIGIN = 0xFFE50000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_0_instruction_cache_MEM_0 : ORIGIN = 0xFFE40000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_1_data_cache_MEM_0 : ORIGIN = 0xFFED0000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_1_instruction_cache_MEM_0 : ORIGIN = 0xFFEC0000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_0_atcm_global_MEM_0 : ORIGIN = 0xFFE00000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_1_atcm_global_MEM_0 : ORIGIN = 0xFFE90000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_0_btcm_global_MEM_0 : ORIGIN = 0xFFE20000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_1_btcm_global_MEM_0 : ORIGIN = 0xFFEB0000, LENGTH = 0x10000'
    if "psv_cortexa72" in machine:
        mem_sec += '\n\tpsv_pmc_ram_psv_pmc_ram : ORIGIN = 0xF2000000, LENGTH = 0x20000'
        mem_sec += '\n\tpsv_r5_0_atcm_global_MEM_0 : ORIGIN = 0xFFE00000, LENGTH = 0x40000'
        mem_sec += '\n\tpsv_r5_1_atcm_global_MEM_0 : ORIGIN = 0xFFE90000, LENGTH = 0x10000'
        mem_sec += '\n\tpsv_r5_1_btcm_global_MEM_0 : ORIGIN = 0xFFEB0000, LENGTH = 0x10000'
    if "psx_cortexa78" in machine:
        mem_sec += '\n\tpsx_pmc_ram : ORIGIN = 0xF2000000, LENGTH = 0x20000'
        mem_sec += '\n\tpsx_r52_0a_atcm_global : ORIGIN = 0xEBA00000, LENGTH = 0x10000'
        mem_sec += '\n\tpsx_r52_0a_btcm_global : ORIGIN = 0xEBA10000, LENGTH = 0x8000'
        mem_sec += '\n\tpsx_r52_0a_ctcm_global : ORIGIN = 0xEBA20000, LENGTH = 0x8000'
        mem_sec += '\n\tpsx_r52_1a_atcm_global : ORIGIN = 0xEBA40000, LENGTH = 0x10000'
        mem_sec += '\n\tpsx_r52_1a_btcm_global : ORIGIN = 0xEBA50000, LENGTH = 0x8000'
        mem_sec += '\n\tpsx_r52_1a_ctcm_global : ORIGIN = 0xEBA60000, LENGTH = 0x8000'
        mem_sec += '\n\tpsx_r52_0b_atcm_global : ORIGIN = 0xEBA80000, LENGTH = 0x10000'
        mem_sec += '\n\tpsx_r52_0b_btcm_global : ORIGIN = 0xEBA90000, LENGTH = 0x8000'
        mem_sec += '\n\tpsx_r52_0b_ctcm_global : ORIGIN = 0xEBAA0000, LENGTH = 0x8000'
        mem_sec += '\n\tpsx_r52_1b_atcm_global : ORIGIN = 0xEBAC0000, LENGTH = 0x10000'
        mem_sec += '\n\tpsx_r52_1b_btcm_global : ORIGIN = 0xEBAD0000, LENGTH = 0x8000'
        mem_sec += '\n\tpsx_r52_1b_ctcm_global : ORIGIN = 0xEBAE0000, LENGTH = 0x8000'
    if "cortexr52" in machine:
        if "psx_cortexr52" in machine:
            mem_sec += '\n\tpsx_pmc_ram : ORIGIN = 0xF2000000, LENGTH = 0x20000'
            mem_sec += '\n\tpsx_r52_tcm_alias : ORIGIN = 0x0, LENGTH = 0x20000'
        else:
            mem_sec += '\n\tpmc_ram : ORIGIN = 0xF2000000, LENGTH = 0x20000'
            mem_sec += '\n\tr52_tcm_alias : ORIGIN = 0x0, LENGTH = 0x20000'
    if openamp_config:
        heap_size = 0x4000
    for key, value in sorted(mem_ranges.items(), key=lambda e: e[1][1], reverse=traverse):
        start,size = value[0], value[1]
        """
        Initial 80 bytes is being used by the linker vectors section in case of Microblaze.
        Adjust the size and start address accordingly.
        """
        if cpu_ip_name == "microblaze" and start < 80:
            start = 80
            size -= start
            if size <= 0x2000 and not user_heap_size:
                heap_size = 0x400
        elif cpu_ip_name == "microblaze" and "lmb_bram" in key :
            start += 80
            size -= 80
        # Handle case where memory starts at vector base address
        elif cpu_ip_name == "microblaze" and mb_reset_addr and start == mb_reset_addr:
            start += 80  # Reserve space for vectors
            size -= 80

        """
        For R5 PSU DDR initial 1MB is reserved for tcm
        Adjust the size and start address accordingly.
        """
        if "psu_ddr" in key and machine == "cortexr5-zynqmp" and start == 0:
            start = 1048576
            size -= start
        if "axi_noc" in key and machine == "cortexr5-versal" and start == 0:
            start = 1048576
            size -= start
        # Use OpenAMP DDR carveout in lieu of system DDR space if DDR OpenAMP carveout is present
        if openamp_elfload_sz and ("psu_ddr" in key or "axi_noc" in key):
                start = openamp_elfload_start
                size = openamp_elfload_sz
                cfd.write("set(%s %s)\n" % (openamp_entry_name, hex(openamp_elfload_start)))

        mem_sec += f'\n\t{lable_names[key]} : ORIGIN = {hex(start)}, LENGTH = {hex(size)}'

    cfd.write(f'set(MEMORY_SECTION "MEMORY\n{{{mem_sec}\n}}")\n')
    if stack_size is not None:
        cfd.write(f'set(STACK_SIZE {hex(stack_size)})\n')
    if heap_size is not None:
        cfd.write(f'set(HEAP_SIZE {hex(heap_size)})\n')



def get_ddr_address(sdt,tgt_node,mem_ranges,match_cpunode,cpu_ip_name,memtest_config,traverse):
    """
    To get the ddr address
    Parameters:
        sdt(obj)                : SDT
        tgt_node(obj)           : Target node obj
        mem_ranges(list)        : List of mem ranges
        match_cpunode(obj)      : Matching Cpu nodes
        cpu_ip_name(str)        : Cpu ip name
        memtest_config(bool)    : True or False
        traverse(bool)          : True or False
    Return:
        default_ddr(str)
        memtest_config(bool)
        mb_reset_addr(str)
    """

    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()
    mem_nodes = []
    symbol_node = ""
    default_ddr = None

    for node in root_sub_nodes:
        try:
            if node.name == "__symbols__":
                symbol_node = node
            device_type = node["device_type"].value
            if "memory" in device_type:
                mem_nodes.append(node)
        except:
           pass

    unique_mem_ip_list = get_unique_memip_list(mem_ranges)
    valid_mem_ips = []
    if (cpu_ip_name == "microblaze" or cpu_ip_name == "microblaze_riscv") and match_cpunode.propval('xlnx,memory-ip-list') != ['']:
        # Get the memory node list
        for node in mem_nodes:
            label_name = get_label(sdt, symbol_node, node)
            if label_name in match_cpunode.propval('xlnx,memory-ip-list'):
                if node.propval('xlnx,ip-name') != ['']:
                    valid_mem_ips.append(node.propval('xlnx,ip-name')[0])
    for key, value in sorted(mem_ranges.items(), key=lambda e: e[1][1], reverse=traverse):
        if default_ddr is None:
            if "axi_emc" in key and len(unique_mem_ip_list) != 1:
                pass
            else:
                if (cpu_ip_name == "microblaze" or cpu_ip_name == "microblaze_riscv") and valid_mem_ips:
                    valid_mem = [mem_ip for mem_ip in valid_mem_ips if key.rsplit('_', 1)[0] in mem_ip]
                    if valid_mem:
                        default_ddr = key
                else:
                    default_ddr = key
    ## To inline with existing tools point default ddr for linker to lower DDR
    lower_ddrs = ["axi_noc", "psu_ddr_0", "ps7_ddr_0"]
    has_ddr = {key:value for key,value in mem_ranges.items() for ddr in lower_ddrs if re.search(ddr, key) and value}
    if has_ddr and not memtest_config and not "microblaze" in cpu_ip_name:
        default_ddr = min(has_ddr, key=lambda k: has_ddr[k])

    mb_reset_addr = None
    if "microblaze" in cpu_ip_name:
        if match_cpunode.propval('xlnx,base-vectors') != ['']:
            if len(match_cpunode.propval('xlnx,base-vectors')) > 1:
                mb_reset_addr = match_cpunode.propval('xlnx,base-vectors')[0] << 32 | match_cpunode.propval('xlnx,base-vectors')[1]
            else:
                mb_reset_addr = match_cpunode.propval('xlnx,base-vectors')[0]
    return default_ddr,memtest_config,mb_reset_addr

def generate_mem_config(mem_ranges,memtest_config,memip_list,default_ddr,cpu_ip_name):
    """
    To generate the memory config
    Parameters:
        mem_ranges(list)        : List of mem ranges
        memtest_config(bool)    : True or False
        memip_list(list)        : Memory list
        default_ddr(int)        : DDR value
        cpu_ip_name(str)        : Cpu IP name
    Return:
        default_ddr(int)
        memip_list(list)

    """
    ## For memory tests configuration default memory should be ocm if available
    if memtest_config:
        has_ocm = [x for x in mem_ranges.keys() if "ocm" in x]
        has_ram = [x for x in mem_ranges.keys() if "_ram" in x]
        has_bram = [x for x in mem_ranges.keys() if "_bram" in x]
        if has_ocm and ("microblaze" not in cpu_ip_name):
            default_ddr = has_ocm[0]
        elif has_ram:
            default_ddr = has_ram[0]
        elif has_bram:
            if len(has_bram) > 1:
                for index, (key, value) in enumerate(mem_ranges.items()):
                    if "lmb_bram" in key:
                        default_ddr = key
                        break
                    if index == 0:
                        size = value[1]
                    if "_bram" in key:
                        if size >= value[1]:
                            size = value[1]
                            default_ddr = key
            else:
                default_ddr = has_bram[0]

        memip_list.insert(0, memip_list.pop(memip_list.index(default_ddr)))

    return default_ddr,memip_list




#!/usr/bin/env python3

#/*
# * Copyright (c) 2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import requests
import getopt
import sys
import os
import json
from pathlib import Path

import lopper.log

# a simple script to replace calls like this:
# curl http://127.0.0.1:5000/domains | python3 -c 'import json,sys;print( json.load(sys.stdin))'

VERSION="0.1-alpha"

def usage():
    prog = os.path.basename(sys.argv[0])
    print('Usage: %s [OPTION] url [<output file>]...' % prog)
    print('  -v, --verbose       enable verbose/debug processing (specify more than once for more verbosity)')
    print('  -j, --json          print unprocessed json response' )
    print('    , --version       output the version and exit')
    print('')
    print(' This is a simple script to replace curl and python on the command line' )
    print('   i.e.: curl http://127.0.0.1:5000/domains | python3 -c \'import json,sys;print( json.load(sys.stdin))\'')
    print('')


def main():
    global verbose
    global json
    global url
    global output
    global force

    url = None
    verbose = 0
    json = False
    output = None
    force = False
    try:
        opts, args = getopt.getopt(sys.argv[1:], "vj", [ "version", "json"])
    except getopt.GetoptError as err:
        lopper.log._error(str(err))
        usage()
        sys.exit(2)

    if opts == [] and args == []:
        usage()
        sys.exit(1)

    for o, a in opts:
        if o in ('-v', "--verbose"):
            verbose = verbose + 1
        elif o in ('--json'):
            json=True
        elif o in ('--version'):
            print( "%s" % VERSION )
            sys.exit(0)
        else:
            assert False, "unhandled option"

    lopper.log._init(__name__)
    lopper.log.init(verbose)

    # any args should be <url> <output file>
    module_name = ""
    module_args= []
    module_args_found = False
    for idx, item in enumerate(args):
        # validate that the system device tree file exists
        if idx == 0:
            url = item
        else:
            if item == "--":
                module_args_found = True

            # the last input is the output file. It can't already exist, unless
            # --force was passed
            if not module_args_found:
                if idx == 1:
                    if output:
                        lopper.log._error("output was already provided via -o")
                        usage()
                        sys.exit(1)
                    else:
                        output = item
                        output_file = Path(output)
                        if output_file.exists():
                            if not force:
                                lopper.log._error( f"output file {output} exists, and -f was not passed" )
                                sys.exit(1)
            else:
                # module arguments
                if not item == "--":
                    if not module_name:
                        module_name = item
                    else:
                        module_args.append( item )

    if not url:
        lopper.log._error("no url was supplied")
        usage()
        sys.exit(1)

    lopper.log._info( f"url: {url}" )

    r = requests.get( url )
    if not r:
        lopper.log._error( f"request failed: {r}" )
    else:
        if json:
            print( r.text )
        else:
            print( r.json() )


if __name__ == "__main__":
    main()
